"""Protobuf interface for message creatia."""

from ansys.api.edb.v1 import arc_data_pb2
from ansys.api.edb.v1.cell_instance_pb2 import (
    CellInstanceCreationMessage,
    CellInstanceParameterOverride,
)
from ansys.api.edb.v1.cell_pb2 import (
    CellCutOutMessage,
    CellFindMessage,
    CellSetTemperatureSettingsMessage,
)
from ansys.api.edb.v1.component_def_pb2 import ComponentDefCreateMessage
from ansys.api.edb.v1.component_group_pb2 import (
    ComponentGroupCreateMessage,
    SetComponentGroupTypeMessage,
)
from ansys.api.edb.v1.die_property_pb2 import (
    DieOrientationMessage,
    DieOrientationPropertyMessage,
    DieTypeMessage,
    DieTypePropertyMessage,
)
from ansys.api.edb.v1.differential_pair_pb2 import (
    DifferentialPairCreationMessage,
    DifferentialPairNetRefsMessage,
)
from ansys.api.edb.v1.edb_messages_pb2 import (
    BoolPropertyMessage,
    ComponentTypeMessage,
    DesignModePropertyMessage,
    DoublePropertyMessage,
    DoublesMessage,
    DoublesPropertyMessage,
    EDBInternalIdMessage,
    EDBObjCollectionMessage,
    EDBObjCollectionPropertyMessage,
    EDBObjMessage,
    EDBObjNameMessage,
    EDBObjPairMessage,
    GetProductPropertyIdsMessage,
    GetProductPropertyMessage,
    HfssExtentInfoMessage,
    HfssExtentMessage,
    IntPropertyMessage,
    PointerPropertyMessage,
    ProductPropertyIdMessage,
    SetProductPropertyMessage,
    StringPairMessage,
    StringPairPropertyMessage,
    StringPropertyMessage,
    StringsMessage,
    StringsPropertyMessage,
    TemperatureSettingsMessage,
    UInt64PropertyMessage,
    ValueMessage,
    ValuePairMessage,
    ValuePairPropertyMessage,
    ValuePropertyMessage,
)
from ansys.api.edb.v1.edge_term_pb2 import (
    EdgeCreationMessage,
    EdgeParamsMessage,
    EdgeTermCreationMessage,
    EdgeTermSetEdgesMessage,
    EdgeType,
    PadEdgeParamsMessage,
    PrimitiveEdgeParamsMessage,
)
from ansys.api.edb.v1.group_pb2 import GroupModifyMemberMessage
from ansys.api.edb.v1.hfss_simulation_settings_pb2 import (
    AdaptiveFrequencyDataMessage,
    AdaptiveMultiFrequencyDataMessage,
    BroadbandFrequencyAdaptiveSolutionMessage,
    MatrixConvergenceDataMessage,
    MultiFrequencyAdaptiveSolutionMessage,
    SingleFrequencyAdaptiveSolutionMessage,
)
from ansys.api.edb.v1.hfss_simulation_setup_pb2 import (
    LengthMeshOperationMessage,
    MeshOperationMessage,
    MeshOpNetLayerInfoMessage,
    SkinDepthMeshOperationMessage,
)
from ansys.api.edb.v1.hierarchy_obj_pb2 import ObjectNameInLayoutMessage
from ansys.api.edb.v1.inst_array_pb2 import InstArrayCreationMessage
from ansys.api.edb.v1.layout_pb2 import LayoutConvertP2VMessage, LayoutExpandedExtentMessage
from ansys.api.edb.v1.material_def_pb2 import MaterialDefPropertiesMessage
from ansys.api.edb.v1.mcad_model_pb2 import *  # noqa
from ansys.api.edb.v1.package_def_pb2 import HeatSinkMessage, SetHeatSinkMessage
from ansys.api.edb.v1.padstack_inst_term_pb2 import (
    PadstackInstTermCreationsMessage,
    PadstackInstTermParamsMessage,
    PadstackInstTermSetParamsMessage,
)
from ansys.api.edb.v1.pin_group_pb2 import (
    PinGroupCreationMessage,
    PinGroupGetUniqueNameMessage,
    PinGroupLookupMessage,
    PinGroupPinsModifyMessage,
)
from ansys.api.edb.v1.pin_group_term_pb2 import (
    PinGroupTermCreationMessage,
    PinGroupTermSetLayerMessage,
    PinGroupTermSetPinGroupMessage,
)
from ansys.api.edb.v1.pin_pair_model_pb2 import PinPairModelRlcPropertyMessage
from ansys.api.edb.v1.point_3d_data_pb2 import *  # noqa
from ansys.api.edb.v1.point_data_pb2 import *  # noqa
from ansys.api.edb.v1.point_term_pb2 import (
    PointTermCreationMessage,
    PointTermParamsMessage,
    PointTermSetParamsMessage,
)
from ansys.api.edb.v1.polygon_data_pb2 import *  # noqa
from ansys.api.edb.v1.port_post_processing_prop_pb2 import PortPostProcessingPropMessage
from ansys.api.edb.v1.refs_pb2 import (
    LayerRefMessage,
    LayerRefPropertyMessage,
    LayerRefsPropertyMessage,
    NetRefMessage,
)
from ansys.api.edb.v1.rlc_pb2 import RlcMessage
from ansys.api.edb.v1.s_parameter_model_pb2 import SParameterModelMessage
from ansys.api.edb.v1.simulation_setup_pb2 import MatrixConvergenceEntryMessage
from ansys.api.edb.v1.spice_model_pb2 import SpiceModelMessage, SpiceModelNewTerminalPinMessage
from ansys.api.edb.v1.structure_3d_pb2 import ClosureMessage, SetClosureMessage
from ansys.api.edb.v1.term_inst_pb2 import TermInstCreationMessage
from ansys.api.edb.v1.term_inst_term_pb2 import (
    TermInstTermCreationMessage,
    TermInstTermSetInstanceMessage,
)
from ansys.api.edb.v1.term_pb2 import (
    TermFindByNameMessage,
    TermGetProductSolversMessage,
    TermSetLayerMessage,
    TermSetParamsMessage,
    TermSetRefMessage,
    TermSetSolverOptionMessage,
)
from ansys.api.edb.v1.transform_pb2 import TransformMessage, TransformPropertyMessage
from ansys.api.edb.v1.via_group_pb2 import (
    ViaGroupCreateWithOutlineMessage,
    ViaGroupCreateWithPrimitivesMessage,
)
from ansys.api.edb.v1.voltage_regulator_pb2 import PowerModuleMessage
from google.protobuf.empty_pb2 import Empty
from google.protobuf.wrappers_pb2 import BoolValue, DoubleValue, FloatValue, Int64Value, StringValue

from ansys.edb.core.simulation_setup.mesh_operation import (
    LengthMeshOperation,
    SkinDepthMeshOperation,
)
from ansys.edb.core.utility import conversions, value


def str_message(s: str):
    """Convert to a ``StringValue`` message."""
    return StringValue(value=s) if s is not None else None


def bool_message(b: bool):
    """Convert to a ``BoolValue`` message."""
    return BoolValue(value=b) if b is not None else None


def int64_message(i: int):
    """Convert to an ``Int64Value`` message."""
    return Int64Value(value=i) if i is not None else None


def float_message(v: float):
    """Convert to a ``FloatValue`` message."""
    return FloatValue(value=v)


def double_message(v: float):
    """Convert to a ``DoubleValue`` message."""
    return DoubleValue(value=v)


def empty_message():
    """Get an empty message."""
    return Empty()


def str_pair_message(pair):
    """Convert to a ``StringPairMessage`` object."""
    return StringPairMessage(first=pair[0], second=pair[1])


def point_message(point):
    """Convert to a ``PointMessage`` object."""
    point = conversions.to_point(point)
    return PointMessage(x=value_message(point.x), y=value_message(point.y))


def point_property_message(target, point):
    """Convert to a ``PointPropertyMessage`` object."""
    return PointPropertyMessage(target=target.msg, point=point_message(point))


def point_data_rotate_message(point, center, angle):
    """Convert to a ``PointRotateMessage`` object."""
    return PointDataRotateMessage(
        point=point_message(point), rotate_center=point_message(center), rotate_angle=angle
    )


def point_data_with_line_message(point, line_start, line_end):
    """Convert to a ``PointDataWithLineMessage`` object."""
    return PointDataWithLineMessage(
        point=point_message(point),
        line_start=point_message(line_start),
        line_end=point_message(line_end),
    )


def box_message(ll, ur):
    """Convert to a ``BoxMessage`` object."""
    return BoxMessage(lower_left=point_message(ll), upper_right=point_message(ur))


def circle_message(center, radius):
    """Convert to a ``CircleMessage`` object."""
    return CircleMessage(center=point_message(center), radius=value_message(radius))


def polygon_data_message(pd):
    """Convert to a ``PolygonDataMessage`` object."""
    return PolygonDataMessage(
        points=[point_message(pt) for pt in pd.points],
        closed=pd.is_closed,
        sense=pd.sense.value,
        holes=[polygon_data_message(h) for h in pd.holes],
    )


def polygon_data_list_message(pds):
    """Convert to a ``PolygonDataListMessage`` object."""
    pds = _as_array(pds)
    return PolygonDataListMessage(polygons=[polygon_data_message(pd) for pd in pds])


def polygon_data_with_tol_message(pd, tol):
    """Convert to a ``PolygonDataWithToleranceMessage`` object."""
    return PolygonDataWithToleranceMessage(polygon=polygon_data_message(pd), tol=tol)


def polygon_data_pair_message(pds1, pds2):
    """Convert to a ``PolygonDataPairMessage`` object."""
    return PolygonDataPairMessage(
        first=polygon_data_list_message(pds1), second=polygon_data_list_message(pds2)
    )


def polygon_data_pair_with_tolerance_message(pd1, pd2, tol):
    """Convert to a ``PolygonDataPairWithToleranceMessage`` object."""
    return PolygonDataPairWithToleranceMessage(
        first=polygon_data_message(pd1), second=polygon_data_message(pd2), tol=tol
    )


def _polygon_data_transform_message_point_value(val, point):
    """Convert to a ``PolygonDataTransformMessage`` object."""
    return PolygonDataTransformMessage.PointValueMessage(point=point_message(point), value=val)


def polygon_data_transform_message(op, pd, *args):
    """Convert to a ``PolygonDataTransformMessage`` object."""
    payload = {}
    if op == "move":
        payload["move"] = point_message(args[0])
    elif op == "rotate":
        payload["rotate"] = _polygon_data_transform_message_point_value(*args)
    elif op == "scale":
        payload["scale"] = _polygon_data_transform_message_point_value(*args)
    elif op == "mirror_x":
        payload["mirror_x"] = args[0]

    return PolygonDataTransformMessage(polygon=polygon_data_message(pd), **payload)


def polygon_data_remove_arc_message(pd, max_chord_error, max_arc_angle, max_points):
    """Convert to a ``PolygonDataRemoveArcMessage`` object."""
    return PolygonDataRemoveArcMessage(
        polygon=polygon_data_message(pd),
        max_chord_error=max_chord_error,
        max_arc_angle=max_arc_angle,
        max_points=max_points,
    )


def polygon_data_with_circle_message(pd, center, radius):
    """Convert to a ``PolygonDataWithCircleMessage`` object."""
    return PolygonDataWithCircleMessage(
        polygon=polygon_data_message(pd), circle=circle_message(center, radius)
    )


def polygon_data_with_point_message(pd, point):
    """Convert to a ``PolygonDataWithPointMessage`` object."""
    return PolygonDataWithPointMessage(target=polygon_data_message(pd), point=point_message(point))


def polygon_data_with_points_message(pd, point=None, polygon=None):
    """Convert to a ``PolygonDataWithPointsMessage`` object."""
    payload = {}
    if point is not None:
        payload["point"] = point_message(point)
    elif polygon is not None:
        payload["polygon"] = polygon_data_message(polygon)

    return PolygonDataWithPointsMessage(target=polygon_data_message(pd), **payload)


def polygon_data_expand_message(pd, offset, tol, round_corner, max_corner_expansion):
    """Convert to a ``PolygonDataExpandMessage`` object."""
    return PolygonDataExpandMessage(
        polygon=polygon_data_message(pd),
        offset=offset,
        tol=tol,
        round_corner=round_corner,
        max_corner_extension=max_corner_expansion,
    )


def polygon_data_get_alpha_shape_message(pd, alpha):
    """Convert to a ``PolygonDataGetAlphaShapeMessage`` object."""
    return PolygonDataGetAlphaShapeMessage(polygon=polygon_data_message(pd), alpha=alpha)


def _arc_rotation_dir(dir):
    if dir.value == "cw":
        return arc_data_pb2.ROTATION_DIRECTION_CW
    elif dir.value == "ccw":
        return arc_data_pb2.ROTATION_DIRECTION_CCW
    elif dir.value == "colinear":
        return arc_data_pb2.ROTATION_DIRECTION_COLINEAR


def arc_message(arc):
    """Convert to an ``ArcMessage`` object."""
    message = arc_data_pb2.ArcMessage(start=point_message(arc.start), end=point_message(arc.end))
    h, opts = arc._height, arc._height_options  # noqa
    if h is not None:
        message.height.value = h
    elif "radius" in opts and "direction" in opts and "is_big" in opts:
        message.radius.radius = opts["radius"]
        message.radius.dir = _arc_rotation_dir(opts["direction"])
        message.radius.is_big = opts["is_big"]
    elif "center" in opts and "direction" in opts:
        message.center.point.CopyFrom(point_message(opts["center"]))
        message.center.dir = _arc_rotation_dir(opts["direction"])
    elif "thru" in opts:
        message.thru.CopyFrom(point_message(opts["thru"]))
    else:
        raise TypeError(
            "`ArcData` instance was not initialized with correct parameters to compute height."
            f"Received '{opts}'"
        )

    return message


def arc_data_two_points(arc1, arc2):
    """Convert to an ``ArcDataTwoArcsMessage`` object."""
    return arc_data_pb2.ArcDataTwoArcsMessage(arc1=arc_message(arc1), arc2=arc_message(arc2))


def int_property_message(target, value):
    """Convert to an ``IntPropertyMessage`` object."""
    return IntPropertyMessage(target=target.msg, value=value)


def uint64_property_message(target, value):
    """Convert to a ``UInt64PropertyMessage`` object."""
    return UInt64PropertyMessage(target=target.msg, value=value)


def bool_property_message(target, value):
    """Convert to a ``BoolPropertyMessage`` object."""
    return BoolPropertyMessage(target=target.msg, value=value)


def string_property_message(target, value):
    """Convert to a ``StringPropertyMessage`` object."""
    return StringPropertyMessage(target=target.msg, value=value)


def string_pair_property_message(target, pair):
    """Convert to a ``StringPairPropertyMessage`` object."""
    return StringPairPropertyMessage(target=target.msg, values=str_pair_message(pair))


def value_property_message(target, value):
    """Convert to a ``ValuePropertyMessage`` object."""
    return ValuePropertyMessage(target=edb_obj_message(target), value=value_message(value))


def pointer_property_message(target, value):
    """Convert to a ``PointerPropertyMessage`` object."""
    return PointerPropertyMessage(target=target.msg, value=value.msg)


def edb_obj_name_message(obj, name):
    """Convert to an ``EDBObjNameMessage`` object."""
    return EDBObjNameMessage(target=edb_obj_message(obj), name=name)


def object_name_in_layout_message(layout, name):
    """Convert to an ``ObjectNameInLayoutMessage`` object."""
    return ObjectNameInLayoutMessage(layout=layout.msg, name=name)


def group_modify_member_message(target, member):
    """Convert to a ``GroupModifyMemberMessage`` object."""
    return GroupModifyMemberMessage(target=target.msg, member=member.msg)


def via_group_create_with_primitives_message(layout, primitives, is_persistent):
    """Convert to a ``ViaGroupCreateWithPrimitivesMessage`` object."""
    return ViaGroupCreateWithPrimitivesMessage(
        layout=layout.msg,
        primitives=edb_obj_collection_message(primitives),
        is_persistent=is_persistent,
    )


def via_group_create_with_outline_message(layout, outline, conductivity_ratio, layer, net=None):
    """Convert to a ``ViaGroupCreateWithOutlineMessage`` object."""
    return ViaGroupCreateWithOutlineMessage(
        layout=layout.msg,
        points=polygon_data_message(outline),
        conductivity_ratio=conductivity_ratio,
        layer=layer_ref_message(layer),
        net=net_ref_message(net),
    )


def cell_instance_creation_message(layout, name, ref):
    """Convert to a ``CellInstanceCreationMessage`` object."""
    return CellInstanceCreationMessage(
        target=object_name_in_layout_message(layout, name), ref=edb_obj_message(ref)
    )


def cell_instance_parameter_override_message(target, param_name, param_value):
    """Convert to a ``CellInstanceParameterOverrideMessage`` object."""
    return CellInstanceParameterOverride(
        target=edb_obj_message(target),
        pname=param_name,
        pval=value_message(param_value),
    )


def inst_array_creation_message(layout, name, ref, orig, xaxis, yaxis, xcount, ycount):
    """Convert to an ``InstArrayCreationMessage`` object."""
    return InstArrayCreationMessage(
        layout=string_property_message(layout, name),
        ref=edb_obj_message(ref),
        orig=point_message(orig),
        xaxis=point_message(xaxis),
        yaxis=point_message(yaxis),
        xcount=value_message(xcount),
        ycount=value_message(ycount),
    )


def component_def_creation_message(db, comp_name, fp):
    """Convert to a ``ComponentDefCreateMessage`` object."""
    if fp is None:
        return ComponentDefCreateMessage(db=db.msg, comp_name=comp_name, fp=None)
    else:
        return ComponentDefCreateMessage(db=db.msg, comp_name=comp_name, fp=fp.msg)


def transform_message(transform):
    """Convert to a ``TransformMessage`` object."""
    if transform is None:
        return None
    else:
        return TransformMessage(
            scale=value_message(transform.scale),
            angle=value_message(transform.rotation),
            mirror=transform.mirror,
            offset_x=value_message(transform.offset_x),
            offset_y=value_message(transform.offset_y),
        )


def transform_property_message(target, transform):
    """Convert to a ``TransformPropertyMessage`` object."""
    return TransformPropertyMessage(target=target.msg, transf=transform_message(transform))


def point3d_message(point3d):
    """Convert to a ``Point3DMessage`` object."""
    if point3d is None:
        return None
    else:
        x, y, z = point3d.x, point3d.y, point3d.z
        return Point3DMessage(x=value_message(x), y=value_message(y), z=value_message(z))


def point_3d_property_message(target, val):
    """Convert to a ``Point3DPropertyMessage`` object."""
    return Point3DPropertyMessage(target=edb_obj_message(target), value=point3d_message(val))


def layout_expanded_extent_message(
    layout, nets, extent, exp, exp_unitless, use_round_corner, num_increments
):
    """Convert to a ``LayoutExpandedExtentMessage`` object."""
    return LayoutExpandedExtentMessage(
        layout=layout.msg,
        nets=edb_obj_collection_message(nets),
        etype=extent.value,
        expansion_factor=exp,
        expansion_factor_unitless=exp_unitless,
        use_round_corner=use_round_corner,
        num_increments=num_increments,
    )


def layout_convert_p2v_message(layout, primitives, is_pins):
    """Convert to a ``LayoutConvertP2VMessage`` message."""
    return LayoutConvertP2VMessage(
        layout=layout.msg, primitives=edb_obj_collection_message(primitives), is_pins=is_pins
    )


def temperature_settings_message(settings):
    """Convert to a ``TemperatureSettingsMessage`` object."""
    return TemperatureSettingsMessage(
        temperature=value_message(settings.temperature),
        include_temp_dependence=settings.include_temp_dependence,
        enable_thermal_feedback=settings.enable_thermal_feedback,
    )


def hfss_extent_message(hfss_val):
    """Convert to an ``ExtentMessage`` object."""
    if type(hfss_val) == float or type(hfss_val) == int:
        val = hfss_val
        absolute = False
    else:
        val, absolute = hfss_val
    return HfssExtentMessage(value=val, absolute=absolute)


def hfss_extent_info_message(hfss_info):
    """Convert to an ``HfssExtentInfoMessage`` object."""
    return HfssExtentInfoMessage(
        use_open_region=hfss_info.use_open_region,
        extent_type=hfss_info.extent_type.value,
        open_region_type=hfss_info.open_region_type.value,
        base_polygon=edb_obj_message(hfss_info.base_polygon),
        dielectric_extent_type=hfss_info.dielectric_extent_type.value,
        dielectric_base_polygon=edb_obj_message(hfss_info.dielectric_base_polygon),
        dielectric=hfss_extent_message(hfss_info.dielectric),
        honor_user_dielectric=hfss_info.honor_user_dielectric,
        airbox_truncate_at_ground=hfss_info.airbox_truncate_at_ground,
        airbox_horizontal=hfss_extent_message(hfss_info.airbox_horizontal),
        airbox_vertical_positive=hfss_extent_message(hfss_info.airbox_vertical_positive),
        airbox_vertical_negative=hfss_extent_message(hfss_info.airbox_vertical_negative),
        sync_airbox_vertical_extent=hfss_info.sync_airbox_vertical_extent,
        is_pml_visible=hfss_info.is_pml_visible,
        operating_frequency=value_message(hfss_info.operating_frequency),
        radiation_level=value_message(hfss_info.radiation_level),
        user_xy_data_extent_for_vertical_expansion=hfss_info.user_xy_data_extent_for_vertical_expansion,
    )


def design_mode_property_message(target, mode):
    """Convert to a ``DesignModePropertyMessage`` object."""
    return DesignModePropertyMessage(target=target.msg, mode=mode.value)


def cell_find_message(database, cell_type, cell_name=None, cell_id=None):
    """Convert to a ``CellFindMessage`` object."""
    if cell_name is not None:
        return CellFindMessage(database=database.msg, type=cell_type.value, name=cell_name)
    elif cell_id is not None:
        return CellFindMessage(database=database.msg, type=cell_type.value, id=cell_id)
    else:
        assert False, "Either name or ID must be provided to find a cell."


def cell_cutout_message(cell, included_nets, clipped_nets, clipping_polygon, clean_clipping):
    """Convert to a ``CellCutOutMessage`` object."""
    return CellCutOutMessage(
        cell=cell.msg,
        included_nets=edb_obj_collection_message(included_nets),
        clipped_nets=edb_obj_collection_message(clipped_nets),
        clipping_polygon=polygon_data_message(clipping_polygon),
        clean_clipping=clean_clipping,
    )


def cell_set_temperature_settings_message(cell, temp_settings):
    """Convert to a ``CellSetTemperatureSettingsMessage`` object."""
    return CellSetTemperatureSettingsMessage(
        cell=cell.msg, temp_settings=temperature_settings_message(temp_settings)
    )


def point_term_params_message(layer, point):
    """Convert to a ``PointTermParamMessage`` object."""
    return PointTermParamsMessage(point=point_message(point), layer=layer_ref_message(layer))


def point_term_set_params_message(term, layer, point):
    """Convert to a ``PointTermSetParamsMessage`` object."""
    return PointTermSetParamsMessage(term=term.msg, params=point_term_params_message(layer, point))


def point_term_creation_message(layout, net, layer, name, point):
    """Convert to a ``PointTermCreationMessage`` object."""
    return PointTermCreationMessage(
        layout=layout.msg,
        net=net_ref_message(net),
        name=name,
        params=point_term_params_message(layer, point),
    )


def padstack_inst_term_params_message(padstack_instance, layer):
    """Convert to a ``PadstackInstTermParamsMessage`` object."""
    return PadstackInstTermParamsMessage(
        padstack_instance=padstack_instance.msg, layer=layer_ref_message(layer)
    )


def padstack_inst_term_creation_message(layout, name, padstack_instance, layer, net, is_ref):
    """Convert to a ``PadstackInstTermCreationMessage`` object."""
    return PadstackInstTermCreationsMessage(
        layout=layout.msg,
        name=name,
        params=padstack_inst_term_params_message(padstack_instance, layer),
        net=net_ref_message(net),
        is_ref=is_ref,
    )


def padstack_inst_term_set_params_message(term, padstack_instance, layer):
    """Convert to a ``PadstackInstTermSetParamsMessage`` object."""
    return PadstackInstTermSetParamsMessage(
        term=term.msg, params=padstack_inst_term_params_message(padstack_instance, layer)
    )


def pin_group_creation_message(layout, name, padstack_instances):
    """Convert to a ``PinGroupCreationMessage`` object."""
    return PinGroupCreationMessage(
        layout=layout.msg, name=name, pins=[pi.msg for pi in padstack_instances]
    )


def pin_group_lookup_message(layout, name):
    """Convert to a ``PinGroupLookupMessage`` object."""
    return PinGroupLookupMessage(layout=layout.msg, name=name)


def pin_group_get_unique_name_message(layout, prefix):
    """Convert to a ``PinGroupGetUniqueNameMessage`` object."""
    return PinGroupGetUniqueNameMessage(layout=layout.msg, prefix=prefix)


def pin_group_pins_modify_message(pin_group, padstack_instances):
    """Convert to a ``PinGroupPinsModifyMessage`` object."""
    return PinGroupPinsModifyMessage(
        pin_group=pin_group.msg, pins=[pi.msg for pi in padstack_instances]
    )


def pin_group_term_creation_message(layout, net_ref, name, pin_group, is_ref):
    """Convert to a ``PinGroupTermCreationMessage`` object."""
    return PinGroupTermCreationMessage(
        layout=layout.msg,
        net=net_ref_message(net_ref),
        name=name,
        pin_group=pin_group.msg,
        is_ref=is_ref,
    )


def pin_group_term_set_pin_group_message(term, pin_group):
    """Convert to a ``PinGroupTermSetPinGroupMessage`` object."""
    return PinGroupTermSetPinGroupMessage(term=term.msg, pin_group=pin_group.msg)


def pin_group_term_set_layer_message(term, layer_ref):
    """Convert to a ``PinGroupTermSetLayerMessage`` object."""
    return PinGroupTermSetLayerMessage(term=term.msg, layer=layer_ref_message(layer_ref))


def edge_creation_message(edge_type, **params):
    """Convert to an ``EdgeCreationMessage`` object."""
    return EdgeCreationMessage(edge_type=edge_type, params=edge_params_message(edge_type, **params))


def edge_params_message(edge_type, **params):
    """Convert to an ``EdgeParamsMessage`` object."""
    if edge_type == EdgeType.PRIMITIVE_EDGE:
        return EdgeParamsMessage(primitve_params=primitive_edge_params_message(**params))
    elif edge_type == EdgeType.PAD_EDGE:
        return EdgeParamsMessage(pad_params=pad_edge_params_message(**params))
    else:
        raise RuntimeError(f"Edge type {edge_type} is not valid.")


def primitive_edge_params_message(primitive, point):
    """Convert to a ``PrimitiveEdgeParamsMessage`` object."""
    return PrimitiveEdgeParamsMessage(primitive=primitive.msg, point=point_message(point))


def pad_edge_params_message(padstack_instance, layer, arc):
    """Convert to a ``PadEdgeParamsMessage`` object."""
    return PadEdgeParamsMessage(
        padstack=padstack_instance.msg,
        layer=layer_ref_message(layer),
        arc=arc_message(arc),
    )


def edge_term_creation_message(layout, net, name, edges, is_ref):
    """Convert to an ``EdgeTermCreationMessage`` object."""
    return EdgeTermCreationMessage(
        layout=layout.msg,
        net=net_ref_message(net),
        name=name,
        edges=[edge.msg for edge in edges],
        is_ref=is_ref,
    )


def edge_term_set_edges_message(terminal, edges):
    """Convert to an ``EdgeTermSetEdgesMessage`` object."""
    return EdgeTermSetEdgesMessage(term=terminal.msg, edges=[edge.msg for edge in edges])


def term_set_params_message(term, **params):
    """Convert to a ``TermSetParamsMessage`` object."""
    payload = {"term": term.msg}

    if "boundary_type" in params:
        payload["boundary_type"] = params.pop("boundary_type")
    if "term_to_ground" in params:
        payload["term_to_ground"] = params.pop("term_to_ground")
    if "hfss_pi_type" in params:
        payload["hfss_pi_type"] = params.pop("hfss_pi_type")
    if "ref_term" in params or "ref_layer" in params:
        ref_term = params.pop("ref_term", ())
        ref_term = ref_term if isinstance(ref_term, tuple) else (ref_term,)
        ref_layer = params.pop("ref_layer", ())
        ref_layer = ref_layer if isinstance(ref_layer, tuple) else (ref_layer,)
        payload["ref"] = term_set_ref_message(ref_term, ref_layer)
    if "is_auto_port" in params:
        payload["is_auto_port"] = bool_message(params.pop("is_auto_port"))
    if "is_circuit_port" in params:
        payload["is_circuit_port"] = bool_message(params.pop("is_circuit_port"))
    if "use_ref_from_hierarchy" in params:
        payload["use_ref_from_hierarchy"] = bool_message(params.pop("use_ref_from_hierarchy"))
    if "name" in params:
        payload["name"] = str_message(params.pop("name"))
    if "impedance" in params:
        payload["impedance"] = value_message(params.pop("impedance"))
    if "source_amplitude" in params:
        payload["source_amplitude"] = value_message(params.pop("source_amplitude"))
    if "source_phase" in params:
        payload["source_phase"] = value_message(params.pop("source_phase"))
    if "s_param_model" in params:
        payload["s_param_model"] = str_message(params.pop("s_param_model"))
    if "rlc" in params:
        payload["rlc"] = rlc_message(params.pop("rlc"))
    if "port_post_processing_prop" in params:
        payload["port_post_processing_prop"] = port_post_processing_prop_message(
            params.pop("port_post_processing_prop")
        )
    if len(params.keys()):
        raise RuntimeError("unknown parameters.")
    return TermSetParamsMessage(**payload)


def term_set_ref_message(term_params, layer_params):
    """Convert to a ``TermSetRefMessage`` object."""
    if len(term_params) and len(layer_params):
        raise RuntimeError("Can only reference either a terminal or a layer.")
    if len(term_params) == 1:
        return TermSetRefMessage(ref_term=edb_obj_message(*term_params))
    if len(layer_params):
        return TermSetRefMessage(ref_layer=term_set_layer_message(*layer_params))


def term_set_layer_message(layer_ref, contexts=None):
    """Convert to a ``TermSetLayerMessage`` object."""
    contexts = [] if contexts is None else contexts
    return TermSetLayerMessage(
        layer=layer_ref_message(layer_ref), contexts=[str_message(ctx) for ctx in contexts]
    )


def term_find_by_name_message(layout, name):
    """Convert to a ``TermFindByNameMessage`` object."""
    return TermFindByNameMessage(layout=layout.msg, name=name)


def term_get_product_solver_message(term, product_id):
    """Convert to a ``TermGetProductSolversMessage`` object."""
    return TermGetProductSolversMessage(term=term.msg, product_id=product_id.value)


def term_set_solver_option_message(term, product_id, name, option):
    """Convert to a ``TermSetSolverOptionMessage`` object."""
    return TermSetSolverOptionMessage(
        term=term.msg, product_id=product_id.value, name=name, option=option
    )


def term_inst_creation_message(layout, net_ref, cell_inst, name):
    """Convert to a ``TermInstCreationMessage`` object."""
    return TermInstCreationMessage(
        layout=layout.msg,
        net=net_ref_message(net_ref),
        cell_inst=cell_inst.msg,
        name=name,
    )


def term_inst_term_creation_message(layout, net_ref, name, term_inst, is_ref):
    """Convert to a ``TermInstTermCreationMessage`` object."""
    return TermInstTermCreationMessage(
        layout=layout.msg,
        net=net_ref_message(net_ref),
        name=name,
        term_inst=term_inst.msg,
        is_ref=is_ref,
    )


def term_inst_term_set_instance_message(term, term_inst):
    """Convert to a ``TermInstTermSetInstanceMessage`` object."""
    return TermInstTermSetInstanceMessage(term=term.msg, term_inst=term_inst.msg)


def edb_obj_message(obj):
    """Convert to an ``EDBObjMessage`` object."""
    if obj is None:
        return None
    elif isinstance(obj, EDBObjMessage):
        return obj
    elif isinstance(obj, int):
        return EDBObjMessage(id=obj)
    else:
        return obj.msg


def edb_obj_collection_message(objs):
    """Convert to an ``EDBObjCollectionMessage``object."""
    return EDBObjCollectionMessage(items=[edb_obj_message(obj) for obj in objs])


def rlc_message(rlc):
    """Convert to an ``RlcMessage`` object."""
    return RlcMessage(
        r=value_message(rlc.r),
        l=value_message(rlc.l),
        c=value_message(rlc.c),
        r_enabled=bool_message(rlc.r_enabled),
        l_enabled=bool_message(rlc.l_enabled),
        c_enabled=bool_message(rlc.c_enabled),
        is_parallel=bool_message(rlc.is_parallel),
    )


def port_post_processing_prop_message(prop):
    """Convert to a ``PortPostProcessingPropMessage`` object."""
    return PortPostProcessingPropMessage(
        voltage_magnitude=value_message(prop.voltage_magnitude),
        voltage_phase=value_message(prop.voltage_phase),
        deembed_length=value_message(prop.deembed_length),
        renormalization_impedance=value_message(prop.renormalization_impedance),
        do_deembed=prop.do_deembed,
        do_deembed_gap_length=prop.do_deembed_gap_l,
        do_renormalize=prop.do_renormalize,
    )


def material_properties_message(**kwargs):
    """Convert to a ``MaterialDefPropertiesMessage`` object."""
    params = {
        key: value_message(value)
        for key, value in kwargs.items()
        if key
        in {
            "permittivity",
            "permeability",
            "conductivity",
            "dielectric_loss_tangent",
            "magnetic_loss_tangent",
            "thermal_conductivity",
            "mass_density",
            "specific_heat",
            "youngs_modulus",
            "poissons_ratio",
            "thermal_expansion_coefficient",
        }
    }
    return MaterialDefPropertiesMessage(**params)


def layer_ref_message(layer):
    """Convert to a ``LayerRefMessage`` object."""
    if layer is None:
        return None
    elif type(layer) == str:
        return LayerRefMessage(name=str_message(layer))
    else:
        return LayerRefMessage(id=edb_obj_message(layer.msg))


def layer_refs_message(layer):
    """Convert to a ``LayerRefsMessage`` object."""
    if layer is None:
        return None
    elif isinstance(layer, list):
        return [layer_ref_message(l) for l in layer]
    else:
        return [layer_ref_message(layer)]


def net_ref_message(net):
    """Convert to a ``NetRefMessage`` object."""
    if type(net) == str:
        return NetRefMessage(name=str_message(net))
    else:
        return NetRefMessage(id=edb_obj_message(net.msg if net is not None else 0))


def _length_mesh_operation_message(mesh_op):
    return LengthMeshOperationMessage(
        max_length=mesh_op.max_length,
        restrict_length=mesh_op.restrict_max_length,
        max_elements=mesh_op.max_elements,
        restrict_max_elements=mesh_op.restrict_max_elements,
    )


def _mesh_op_skin_depth_message(mesh_op):
    return SkinDepthMeshOperationMessage(
        skin_depth=mesh_op.skin_depth,
        surface_triangle_length=mesh_op.surface_triangle_length,
        num_layers=mesh_op.number_of_layers,
        max_elements=mesh_op.max_elements,
        restrict_max_elements=mesh_op.restrict_max_elements,
    )


def mesh_operation_message(mesh_op):
    """Convert to a ``MeshOperationMessage`` object."""
    mesh_op_msg = MeshOperationMessage(
        name=mesh_op.name,
        enabled=mesh_op.enabled,
        refine_inside=mesh_op.refine_inside,
        mesh_region=mesh_op.mesh_region,
        net_layer_info=[_mesh_op_net_layer_message(*nl) for nl in mesh_op.net_layer_info],
        solve_inside=mesh_op.solve_inside,
    )
    if isinstance(mesh_op, LengthMeshOperation):
        mesh_op_msg.length_mesh_op.CopyFrom(_length_mesh_operation_message(mesh_op))
    elif isinstance(mesh_op, SkinDepthMeshOperation):
        mesh_op_msg.skin_depth_mesh_op.CopyFrom(_mesh_op_skin_depth_message(mesh_op))
    return mesh_op_msg


def _mesh_op_net_layer_message(net, layer, is_sheet):
    return MeshOpNetLayerInfoMessage(net=net, layer=layer, is_sheet=is_sheet)


def value_message(val):
    """Convert data into a ``ValueMessage`` object.

    Parameters
    ----------
    val : str, int, float, complex, ansys.edb.core.utility.value.Value, ValueMessage

    Returns
    -------
    ValueMessage
    """
    if isinstance(val, value.Value):
        return val.msg
    if isinstance(val, ValueMessage):
        return val

    msg = ValueMessage()
    if isinstance(val, str):
        msg.text = val
        msg.variable_owner.id = 0
    elif isinstance(val, float) or isinstance(val, int):
        msg.constant.real = val
        msg.constant.imag = 0
    elif isinstance(val, complex):
        msg.constant.real = val.real
        msg.constant.imag = val.imag
    else:
        raise TypeError(f"Invalid Value. Received {val}")
    return msg


def product_property_id_message(prod_id, att_id):
    """Convert to a ``ProductPropertyIdMessage`` object."""
    return ProductPropertyIdMessage(product_id=prod_id.value, attribute_id=att_id)


def set_product_property_message(obj, prod_id, att_id, value):
    """Convert to a ``SetProductPropertyMessage`` object."""
    return SetProductPropertyMessage(
        edb_obj=obj.msg,
        product_prop_id=product_property_id_message(prod_id, att_id),
        property_value=value,
    )


def get_product_property_message(obj, prod_id, att_id):
    """Convert to a ``GetProductPropertyMessage`` object."""
    return GetProductPropertyMessage(
        edb_obj=obj.msg, product_prop_id=product_property_id_message(prod_id, att_id)
    )


def get_product_property_ids_message(obj, prod_id):
    """Convert to a ``GetProductPropertyIdsMessage`` object."""
    return GetProductPropertyIdsMessage(edb_obj=obj.msg, product_id=prod_id.value)


def edb_internal_id_message(id):
    """Convert to an ``EDBInternalIdMessage`` object."""
    return EDBInternalIdMessage(id=id)


def component_group_create_message(layout, name, comp_name):
    """Convert to a ``ComponentGroupCreateMessage`` object."""
    return ComponentGroupCreateMessage(
        target=object_name_in_layout_message(layout, name), comp=comp_name
    )


def set_component_group_type_message(obj, comp_type):
    """Convert to a ``SetComponentGroupTypeMessage`` object."""
    return SetComponentGroupTypeMessage(
        target=obj.msg, comp_type=ComponentTypeMessage(comp_type=comp_type.value)
    )


def set_closure_message(obj, closure_type):
    """Convert to a ``SetClosureMessage`` object."""
    return SetClosureMessage(
        target=obj.msg, closure_type=ClosureMessage(closure_type=closure_type.value)
    )


def strings_message(strings):
    """Convert to a ``StringsMessage`` object."""
    return StringsMessage(strings=strings)


def strings_property_message(target, value):
    """Convert to a ``StringPropertyMessage`` object."""
    return StringsPropertyMessage(edb_obj=target.msg, strings=value)


def edb_obj_pair_message(edb_obj_0, edb_obj_1):
    """Convert to an ``EDBObjPairMessage`` object."""
    return EDBObjPairMessage(edb_obj_0=edb_obj_0.msg, edb_obj_1=edb_obj_1.msg)


def layer_ref_property_message(edb_obj, layer_ref):
    """Convert to a ``LayerRefPropertyMessage`` object."""
    return LayerRefPropertyMessage(edb_obj=edb_obj.msg, layer_ref=layer_ref_message(layer_ref))


def layer_refs_property_message(edb_obj, layer_refs):
    """Convert to a ``LayerRefsPropertyMessage`` object."""
    return LayerRefsPropertyMessage(edb_obj=edb_obj.msg, layer_refs=layer_refs_message(layer_refs))


def double_property_message(edb_obj, double):
    """Convert to a ``DoublePropertyMessage`` object."""
    return DoublePropertyMessage(target=edb_obj.msg, value=double)


def differential_pair_creation_message(layout, name, pos_net, neg_net):
    """Convert to a ``DifferentialPairCreationMessage`` object."""
    return DifferentialPairCreationMessage(
        layout=edb_obj_message(layout),
        name=name,
        positive_net=net_ref_message(pos_net),
        negative_net=net_ref_message(neg_net),
    )


def differential_pair_net_refs_message(dp, pos_net, neg_net):
    """Convert to a ``DifferentialPairNetRefsMessage`` object."""
    return DifferentialPairNetRefsMessage(
        dp=edb_obj_message(dp),
        positive_net=net_ref_message(pos_net),
        negative_net=net_ref_message(neg_net),
    )


def mcad_model_creation_message(connectable, layout, filename):
    """Convert to an ``McadModelCreationMessage`` object."""
    if connectable is not None:
        param = McadModelCreationMessage.WithConnObj(obj=edb_obj_message(connectable))
        return McadModelCreationMessage(conn_obj=param)
    elif layout is not None and filename is not None and len(filename) > 0:
        param = McadModelCreationMessage.WithLayout(obj=edb_obj_message(layout), file_name=filename)
        return McadModelCreationMessage(layout=param)
    else:
        raise TypeError("Either a connectable object or layout and filename must be provided.")


def mcad_model_hfss_creation_message(connectable, layout, filename, design):
    """Convert to an ``McadModelHfssCreationMessage`` object."""
    if connectable is not None:
        param = McadModelHfssCreationMessage.WithConnObj(obj=edb_obj_message(connectable))
        return McadModelCreationMessage(conn_obj=param)
    elif layout is not None and filename is not None and len(filename) > 0 and len(design) > 0:
        param = McadModelHfssCreationMessage.WithLayout(
            obj=edb_obj_message(layout), file_name=filename, design=design
        )
        return McadModelCreationMessage(layout=param)
    else:
        raise TypeError(
            "Either a connectable object or layout, filename, and design must be provided."
        )


def mcad_model_rotation_message(axis_from, axis_to, angle):
    """Convert to an ``McadModelRotationMessage`` object."""
    return McadModelRotationMessage(
        axis_from=point3d_message(axis_from), axis_to=point3d_message(axis_to), angle=angle
    )


def mcad_model_set_rotation_message(mcad_model, axis_from, axis_to, angle):
    """Convert to an ``McadModelSetRotationMessage`` object."""
    return McadModelSetRotationMessage(
        model=edb_obj_message(mcad_model),
        rotation=mcad_model_rotation_message(axis_from, axis_to, angle),
    )


def mcad_model_bool_message(mcad_model, index, value):
    """Convert to an ``McadModelBoolMessage`` object."""
    return McadModelBoolMessage(model=edb_obj_message(mcad_model), index=index, value=value)


def mcad_model_string_message(mcad_model, index, value):
    """Convert to an ``McadModelStringMessage`` object."""
    return McadModelStringMessage(model=edb_obj_message(mcad_model), index=index, value=value)


def set_die_type_message(obj, die_type):
    """Convert to a ``DieTypePropertyMessage`` object."""
    return DieTypePropertyMessage(target=obj.msg, die_type=DieTypeMessage(die_type=die_type.value))


def set_die_orientation_message(obj, die_orientation):
    """Convert to a ``DieOrientationPropertyMessage`` object."""
    return DieOrientationPropertyMessage(
        target=obj.msg, die_orientation=DieOrientationMessage(die_orientation=die_orientation.value)
    )


def value_pair_message(val1, val2):
    """Convert to a ``ValuePairMessage`` object."""
    return ValuePairMessage(val1=value_message(val1), val2=value_message(val2))


def value_pair_property_message(target, val1, val2):
    """Convert to a ``ValuePairPropertyMessage`` object."""
    return ValuePairPropertyMessage(
        target=edb_obj_message(target), values=value_pair_message(val1, val2)
    )


def points_message(points):
    """Convert to a ``PointsMessage`` object."""
    return PointsMessage(points=[point_message(point) for point in points])


def points_property_message(target, points):
    """Convert to a ``PointsPropertyMessage`` object."""
    return PointsPropertyMessage(
        target=edb_obj_message(target), points=[point_message(point) for point in points]
    )


def polygon_data_property_message(obj, polygon):
    """Convert to a ``PolygonDataPropertyMessage`` object."""
    return PolygonDataPropertyMessage(target=obj.msg, value=polygon_data_message(polygon))


def heat_sink_message(heat_sink):
    """Convert to a ``HeatSinkMessage`` object."""
    return HeatSinkMessage(
        thickness=value_message(heat_sink.fin_thickness),
        spacing=value_message(heat_sink.fin_spacing),
        base_height=value_message(heat_sink.fin_base_height),
        height=value_message(heat_sink.fin_height),
        orientation=heat_sink.fin_orientation.value,
    )


def set_heat_sink_message(target, heat_sink):
    """Convert to a ``SetHeatSinkMessage`` object."""
    return SetHeatSinkMessage(target=edb_obj_message(target), value=heat_sink_message(heat_sink))


def pin_pair_model_rlc_message(model, pin_pair, rlc):
    """Convert to a ``PinPairModelRlcPropertyMessage`` object."""
    return PinPairModelRlcPropertyMessage(
        model=edb_obj_message(model), pins=str_pair_message(pin_pair), rlc=rlc_message(rlc)
    )


def point_pair_message(point_pair):
    """Convert to a ``PointPairMessage`` object."""
    return PointPairMessage(
        point_0=point_message(point_pair[0]), point_1=point_message(point_pair[1])
    )


def sparameter_model_message(name, ref_net):
    """Convert to a ``SParameterModelMessage`` object."""
    return SParameterModelMessage(name=name, ref_net=ref_net)


def power_module_message(power_module):
    """Convert to a ``PowerModuleMessage`` object."""
    return PowerModuleMessage(
        comp_group_name=power_module.comp_group_name,
        pos_output_terminal=power_module.pos_output_terminal,
        neg_output_terminal=power_module.neg_output_terminal,
        relative_strength=value_message(power_module.relative_strength),
        active=power_module.active,
        needs_sync=power_module.needs_sync,
    )


def point_pair_property_message(target, point_pair):
    """Convert to a ``PointPairPropertyMessage`` object."""
    return PointPairPropertyMessage(
        target=edb_obj_message(target), point_pair=point_pair_message(point_pair)
    )


def spice_model_message(name, path, sub_circuit):
    """Convert to a ``SpiceModelMessage`` object."""
    return SpiceModelMessage(name=name, path=path, sub_ckt=sub_circuit)


def spice_model_net_terminal_pin_message(model, terminal, pin):
    """Convert to a ``SpiceModelNewTerminalPinMessage`` object."""
    return SpiceModelNewTerminalPinMessage(
        target=edb_obj_message(model), terminal=terminal, pin=pin
    )


def doubles_message(doubles):
    """Convert to a ``DoublesMessage`` object."""
    return DoublesMessage(doubles=doubles)


def doubles_property_message(edb_obj, doubles):
    """Convert to a ``DoublesPropertyMessage`` object."""
    return DoublesPropertyMessage(edb_obj=edb_obj_message(edb_obj), doubles=doubles)


def cpos_3d_message(point3d):
    """Convert to a ``CPos3DMessage`` object."""
    if point3d is None:
        return None
    else:
        point3d_converted = conversions.to_point3d(point3d)
        x, y, z = point3d_converted.x.double, point3d_converted.y.double, point3d_converted.z.double
        return CPos3DMessage(x=x, y=y, z=z)


def cpos_3d_property_message(target, value):
    """Convert to a ``CPos3DPropertyMessage`` object."""
    return CPos3DPropertyMessage(target=edb_obj_message(target), value=cpos_3d_message(value))


def cpos_3d_pair_message(x, y):
    """Convert to a ``CPos3DPairMessage`` object."""
    return CPos3DPairMessage(x=cpos_3d_message(x), y=cpos_3d_message(y))


def cpos_3d_triple_message(x, y, z):
    """Convert to a ``CPos3DTripleMessage`` object."""
    return CPos3DTripleMessage(x=cpos_3d_message(x), y=cpos_3d_message(y), z=cpos_3d_message(z))


def cpos_3d_double_message(pos, value):
    """Convert to a ``CPos3DDoubleMessage`` object."""
    return CPos3DDoubleMessage(pos=cpos_3d_message(pos), value=value)


def mx_convergence_entry_msg_list(mx_convergence_entry_list):
    """Convert to a list ``of MatrixConvergenceEntryMessage`` objects."""
    mx_entry_msgs = []
    for mx_entry in mx_convergence_entry_list:
        mx_entry_msgs.append(
            MatrixConvergenceEntryMessage(
                port_1=mx_entry.port_1_name,
                port_2=mx_entry.port_2_name,
                mag_limit=mx_entry.mag_limit,
                phase_limit=mx_entry.phase_limit,
            )
        )
    return mx_entry_msgs


def mx_convergence_data_msg(mx_data):
    """Convert to MatrixConvergenceDataMessage."""
    return MatrixConvergenceDataMessage(
        all_are_constant=mx_data.all_constant,
        all_diag_are_constant=mx_data.all_diag_constant,
        all_off_diag_are_constant=mx_data.all_off_diag_constant,
        mag_min_threashold=mx_data.mag_min_threshold,
        entries=mx_convergence_entry_msg_list(mx_data.entry_list),
    )


def base_adaptive_frequency_solution_msg(adaptive_frequency, max_delta):
    """Convert to an ``AdaptiveFrequencyDataMessage`` object."""
    return AdaptiveFrequencyDataMessage(adaptive_frequency=adaptive_frequency, max_delta=max_delta)


def single_frequency_adaptive_solution_msg(single_freq_adapt_sol):
    """Convert to a ``SingleFrequencyAdaptiveSolutionMessage`` object."""
    return SingleFrequencyAdaptiveSolutionMessage(
        adaptive_frequency=base_adaptive_frequency_solution_msg(
            single_freq_adapt_sol.adaptive_frequency, single_freq_adapt_sol.max_delta
        ),
        max_passes=single_freq_adapt_sol.max_passes,
        matrix_conv_data=mx_convergence_data_msg(single_freq_adapt_sol.mx_conv_data),
        use_matrix_conv_data=single_freq_adapt_sol.use_mx_conv_data,
    )


def multi_adaptive_freq_to_msg(multi_adaptive_freq):
    """Convert to an ``AdaptiveMultiFrequencyDataMessage`` object."""
    return AdaptiveMultiFrequencyDataMessage(
        adaptive_frequency=base_adaptive_frequency_solution_msg(
            multi_adaptive_freq.adaptive_frequency, multi_adaptive_freq.max_delta
        ),
        output_variables=multi_adaptive_freq.output_variables,
    )


def multi_frequency_adaptive_solution_msg(multi_freq_adapt_sol):
    """Convert to a ``MultiFrequencyAdaptiveSolutionMessage`` object."""
    return MultiFrequencyAdaptiveSolutionMessage(
        adaptive_frequencies=[
            multi_adaptive_freq_to_msg(multi_adaptive_freq)
            for multi_adaptive_freq in multi_freq_adapt_sol.adaptive_frequencies
        ],
        max_passes=multi_freq_adapt_sol.max_passes,
    )


def broadband_solution_msg(broadband_adapt_sol):
    """Convert to a ``BroadbandFrequencyAdaptiveSolutionMessage`` object."""
    return BroadbandFrequencyAdaptiveSolutionMessage(
        max_delta=broadband_adapt_sol.max_delta,
        max_passes=broadband_adapt_sol.max_num_passes,
        low_frequency=broadband_adapt_sol.low_frequency,
        high_frequency=broadband_adapt_sol.high_frequency,
    )


def edb_obj_collection_property_message(target, edb_objs):
    """Convert to a ``EDBObjCollectionPropertyMessage`` object."""
    return EDBObjCollectionPropertyMessage(
        target=target.msg, edb_objs=edb_obj_collection_message(edb_objs)
    )


def _as_array(array_or_item):
    try:
        iter(array_or_item)
        return array_or_item
    except TypeError:
        return [array_or_item]



"""HFSS extent information."""
from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike
    from ansys.edb.core.primitive.primitive import Primitive

from enum import Enum

import ansys.api.edb.v1.edb_defs_pb2 as edb_defs_pb2

from ansys.edb.core.utility import conversions
from ansys.edb.core.utility.value import Value


class HFSSExtentInfoType(Enum):
    """Provides an enum representing available HFSS extent information types."""

    BOUNDING_BOX = edb_defs_pb2.HFSS_EXTENT_BOUNDING_BOX
    CONFORMING = edb_defs_pb2.HFSS_EXTENT_CONFIRMING
    CONVEX_HUL = edb_defs_pb2.HFSS_EXTENT_CONVEX_HULL
    POLYGON = edb_defs_pb2.HFSS_EXTENT_POLYGON


class OpenRegionType(Enum):
    """Provides an enum representing open region types."""

    RADIATION = edb_defs_pb2.HFSS_EXTENT_RADIATION
    PML = edb_defs_pb2.HFSS_EXTENT_PML


class HfssExtentInfo:
    """Provides HFSS extent information.

    Attributes
    ----------
        use_open_region : bool, default: ``True``
            Whether an open region is used.
        extent_type : .HFSSExtentInfoType, default: BOUNDING_BOX
            Extent type.
        open_region_type : .OpenRegionType, default: RADIATION
            Open region type.
        base_polygon : .Primitive, default: None
            Polygon to use if the extent is the ``Polygon`` type.
        dielectric_extent_type : .HFSSExtentInfoType, default: BOUNDING_BOX
            Dielectric extent type.
        dielectric_base_polygon : .Primitive, default: None
            Polygon to use if dielectric extent is the ``Polygon`` type.
        dielectric : (float, bool), default: (0, True)
            Dielectric extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        honor_user_dielectric : bool, default: ``True``
            Whether to honor a user-defined dielectric primitive when calculating the dielectric
            extent.
        airbox_truncate_at_ground : bool, default: ``False``
            Whether to truncate the airbox at the ground layers.
        airbox_horizontal : (float, bool), default: (0.15, True)
            Airbox horizontal extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        airbox_vertical_positive : (float, bool), default: (0.15, True)
            Airbox positive vertical extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        airbox_vertical_negative : (float, bool), default: (0.15, True)
            Airbox negative vertical extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        sync_airbox_vertical_extent : bool, default: ``True``
            Whether to synchronize the airbox positive and negative vertical extent.
        is_pml_visible : bool, default: ``False``
            Whether to check to see if PML boxes are to be rendered.
        operating_frequency : :term:`ValueLike`, default: ``"5GHz"``
            PML operating frequency.
        radiation_level : .Value, default: 0
            PML radiation level for calculating the thickness of the boundary.
        user_xy_data_extent_for_vertical_expansion : bool, default: ``True``
            Whether to retain the old behavior for the vertical expansion of the airbox.
            The default is ``True``, in which case the vertical extent is calculated from
            the XY data extent.
    """

    def __init__(
        self,
        use_open_region: bool = True,
        extent_type: HFSSExtentInfoType = HFSSExtentInfoType.BOUNDING_BOX,
        open_region_type: OpenRegionType = OpenRegionType.RADIATION,
        base_polygon: Primitive = None,
        dielectric_extent_type: HFSSExtentInfoType = HFSSExtentInfoType.BOUNDING_BOX,
        dielectric_base_polygon: Primitive = None,
        dielectric: Tuple[float, bool] = (0, True),
        honor_user_dielectric: bool = True,
        airbox_truncate_at_ground: bool = False,
        airbox_horizontal: Tuple[float, bool] = (0.15, True),
        airbox_vertical_positive: Tuple[float, bool] = (0.15, True),
        airbox_vertical_negative: Tuple[float, bool] = (0.15, True),
        sync_airbox_vertical_extent: bool = True,
        is_pml_visible: bool = False,
        operating_frequency: ValueLike = "5GHz",
        radiation_level: Value = Value(0),
        user_xy_data_extent_for_vertical_expansion: bool = True,
    ):
        """Create an HFSS extent information object.

        Parameters
        ----------
        use_open_region : bool, default: ``True``
            Whether an open region is used.
        extent_type : .HFSSExtentInfoType, default: BOUNDING_BOX
            Extent type.
        open_region_type: .OpenRegionType, default: RADIATION
            Open region type.
        base_polygon : .Primitive, default: None
            Polygon to use if the extent is the ``Polygon`` type.
        dielectric_extent_type : .HFSSExtentInfoType, default: BOUNDING_BOX
            Dielectric extent type.
        dielectric_base_polygon : .Primitive, default: None
            Polygon to use if dielectric extent is the ``Polygon`` type.
        dielectric : (float, bool), default: (0, True)
            Dielectric extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        honor_user_dielectric : bool, default: ``True``
            Whether to honor a user-defined dielectric primitive when calculating the dielectric
            extent.
        airbox_truncate_at_ground : bool, default: ``False``
            Whether to truncate the airbox at the ground layers.
        airbox_horizontal : (float, bool), default: (0.15, True)
            Airbox horizontal extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        airbox_vertical_positive : (float, bool), default: (0.15, True)
            Airbox positive vertical extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        airbox_vertical_negative : (float, bool), default: (0.15, True)
            Airbox negative vertical extent size. The first parameter is the value.
            The second parameter is a Boolean indicating if the value is a multiple.
        sync_airbox_vertical_extent : bool, default: ``True``
            Whether to synchronize the airbox positive and negative vertical extent.
        is_pml_visible : bool, default: ``False``
            Whether to check to see if PML boxes are to be rendered.
        operating_frequency : :term:`ValueLike`, default: ``"5GHz"``
            PML operating frequency.
        radiation_level : .Value, default: 0
            PML radiation level for calculating the thickness of the boundary.
        user_xy_data_extent_for_vertical_expansion : bool, default: ``True``
            Whether to retain the old behavior for the vertical expansion of the airbox.
            The default is ``True``, in which case the vertical extent is calculated from
            the XY data extent.
        """
        self.use_open_region = use_open_region
        self.extent_type = extent_type
        self.open_region_type = open_region_type
        self.base_polygon = base_polygon
        self.dielectric_extent_type = dielectric_extent_type
        self.dielectric_base_polygon = dielectric_base_polygon
        self.dielectric = dielectric
        self.honor_user_dielectric = honor_user_dielectric
        self.airbox_truncate_at_ground = airbox_truncate_at_ground
        self.airbox_horizontal = airbox_horizontal
        self.airbox_vertical_positive = airbox_vertical_positive
        self.airbox_vertical_negative = airbox_vertical_negative
        self.sync_airbox_vertical_extent = sync_airbox_vertical_extent
        self.is_pml_visible = is_pml_visible
        self.operating_frequency = conversions.to_value(operating_frequency)
        self.radiation_level = radiation_level
        self.user_xy_data_extent_for_vertical_expansion = user_xy_data_extent_for_vertical_expansion



"""This module performs conversions from arbitrary user input to explicit types."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike, PointLike, Point3DLike

from ansys.api.edb.v1.edb_messages_pb2 import ValueMessage

from ansys.edb.core.geometry import point3d_data, point_data
from ansys.edb.core.utility import value


def to_value(val: ValueLike) -> value.Value:
    """Take a value implicitly convertible to a :class:`.Value` type and return as a :class:`.Value` type.

    Parameters
    ----------
    val : :term:`ValueLike`

    Returns
    -------
    .Value
    """
    if isinstance(val, value.Value):
        return val
    elif type(val) in [int, float, complex, str, ValueMessage]:
        return value.Value(val)
    else:
        raise TypeError(
            f"Value-like objects must be either of type Value or int/float/complex/str. - Received '{val}'"
        )


def to_point(val: PointLike) -> point_data.PointData:
    """Take a value implicitly convertible to a :class:`.PointData` type and return as a :class:`.PointData` type.

    Parameters
    ----------
    val : :term:`Point2DLike`

    Returns
    -------
    .PointData
    """
    if isinstance(val, point_data.PointData):
        return val
    try:
        if len(val) == 2:
            return point_data.PointData(val)
    except TypeError:
        return point_data.PointData(val)

    raise TypeError(
        "Point-like objects must be either of type PointData or a list/tuple containing (start, end) or (arc_height)."
    )


def to_point3d(val: Point3DLike) -> point3d_data.Point3DData:
    """Take a value implicitly convertible to a :class:`.Point3DData` type and return as a :class:`.Point3DData` type.

    Parameters
    ----------
    val : :term:`Point3DLike`

    Returns
    -------
    .Point3DData
    """
    if isinstance(val, point3d_data.Point3DData):
        return val
    if len(val) == 3:
        return point3d_data.Point3DData(val[0], val[1], val[2])



"""Netlist model."""
from ansys.edb.core.hierarchy.model import Model
from ansys.edb.core.inner import messages
from ansys.edb.core.session import NetlistModelServiceStub, StubAccessor, StubType


class NetlistModel(Model):
    """Represents a netlist model object."""

    __stub: NetlistModelServiceStub = StubAccessor(StubType.netlist_model)

    @classmethod
    def create(cls, name):
        """Create a netlist model.

        Parameters
        ----------
        name : str
            Name of the netlist model.

        Returns
        -------
        NetlistModel
            Netlist model created.
        """
        return cls(cls.__stub.Create(messages.str_message(name)))

    @property
    def netlist(self):
        """:obj:`str`: Netlist name."""
        return self.__stub.GetNetlist(messages.edb_obj_message(self)).value

    @netlist.setter
    def netlist(self, name):
        self.__stub.SetNetlist(messages.string_property_message(self, name))



"""Simulation Settings."""
from enum import Enum

import ansys.api.edb.v1.simulation_settings_pb2 as pb

from ansys.edb.core.inner import messages
from ansys.edb.core.session import (
    AdvancedMeshingSettingsServiceStub,
    AdvancedSettingsServiceStub,
    SettingsOptionsServiceStub,
    SimulationSettingsServiceStub,
    SolverSettingsServiceStub,
    StubAccessor,
    StubType,
)


class ViaStyle(Enum):
    """Enum representing via modeling styles."""

    WIREBOND = pb.WIREBOND
    RIBBON = pb.RIBBON
    MESH = pb.MESH
    FIELD = pb.FIELD
    NUM_VIA_STYLE = pb.NUM_VIA_STYLE


class ModelType(Enum):
    """Enum representing defeature model types."""

    GENERAL_MODEL = pb.GENERAL_MODEL
    IC_MODEL = pb.IC_MODEL


class SimulationSettingsBase:
    """Internal base class for simulation settings."""

    def __init__(self, sim_setup):
        """Create a SimulationSettingsBase object."""
        self._sim_setup = sim_setup

    @property
    def msg(self):
        """:obj:`EDBObjMessage`: Protobuf message that represents this object's ID.

        This property is read-only.
        """
        return self._sim_setup.msg


class SimulationSettings(SimulationSettingsBase):
    """Class representing base simulation settings."""

    __stub: SimulationSettingsServiceStub = StubAccessor(StubType.sim_settings)

    @property
    def enabled(self):
        """:obj:`bool`: Flag indicating if the simulation setup is enabled."""
        return self.__stub.GetEnabled(self.msg).value

    @enabled.setter
    def enabled(self, enabled):
        self.__stub.SetEnabled(messages.bool_property_message(self, enabled))


class SettingsOptions(SimulationSettingsBase):
    """Class representing options for base simulation settings."""

    __stub: SettingsOptionsServiceStub = StubAccessor(StubType.sim_settings_options)

    @property
    def do_lambda_refine(self):
        """:obj:`bool`: Flag indicating if lambda refinement is used during meshing."""
        return self.__stub.GetDoLamdaRefineFlag(self.msg).value

    @do_lambda_refine.setter
    def do_lambda_refine(self, do_lambda_refine):
        self.__stub.SetDoLamdaRefineFlag(messages.bool_property_message(self, do_lambda_refine))

    @property
    def lambda_target(self):
        """:obj:`float`: Target lambda value for lambda refinement."""
        return self.__stub.GetLamdaTarget(self.msg).value

    @lambda_target.setter
    def lambda_target(self, lambda_target):
        self.__stub.SetLamdaTarget(messages.double_property_message(self, lambda_target))

    @property
    def mesh_size_factor(self):
        """:obj:`float`: Mesh size factor used for lambda refinement."""
        return self.__stub.GetMeshSizefactor(self.msg).value

    @mesh_size_factor.setter
    def mesh_size_factor(self, mesh_size_factor):
        self.__stub.SetMeshSizefactor(messages.double_property_message(self, mesh_size_factor))

    @property
    def use_default_lambda_value(self):
        """:obj:`bool`: Flag indicating if the default lambda target value is used."""
        return self.__stub.GetLamdaTarget(self.msg).value

    @use_default_lambda_value.setter
    def use_default_lambda_value(self, use_default_value):
        self.__stub.SetDoLamdaRefineFlag(messages.bool_property_message(self, use_default_value))


class AdvancedSettings(SimulationSettingsBase):
    """Class representing base advanced simulation settings."""

    __stub: AdvancedSettingsServiceStub = StubAccessor(StubType.advanced_sim_settings)

    @property
    def union_polygons(self):
        """:obj:`bool`: Flag indicating if polygons are unioned before meshing."""
        return self.__stub.GetUnionPolygons(self.msg).value

    @union_polygons.setter
    def union_polygons(self, union_polygons):
        self.__stub.SetUnionPolygons(messages.bool_property_message(self, union_polygons))

    @property
    def use_defeature(self):
        """:obj:`bool`: Flag indicating if polygon defeaturing is performed."""
        return self.__stub.GetUseDefeature(self.msg).value

    @use_defeature.setter
    def use_defeature(self, use_defeature):
        self.__stub.SetUseDefeature(messages.bool_property_message(self, use_defeature))

    @property
    def use_defeature_absolute_length(self):
        """:obj:`bool`: Flag indicating if absolute length or extent ratio is used when defeaturing polygons."""
        return self.__stub.GetUseDefeatureAbsoluteLength(self.msg).value

    @use_defeature_absolute_length.setter
    def use_defeature_absolute_length(self, use_defeature_absolute_length):
        self.__stub.SetUseDefeatureAbsoluteLength(
            messages.bool_property_message(self, use_defeature_absolute_length)
        )

    @property
    def remove_floating_geometry(self):
        """:obj:`bool`: Flag indicating if a geometry not connected to any other geometry is removed."""
        return self.__stub.GetRemoveFloatingGeometry(self.msg).value

    @remove_floating_geometry.setter
    def remove_floating_geometry(self, remove_floating_geometry):
        self.__stub.SetRemoveFloatingGeometry(
            messages.bool_property_message(self, remove_floating_geometry)
        )

    @property
    def healing_option(self):
        """:obj:`int`: Enable/disable healing of mis-aligned points and edges."""
        return self.__stub.GetHealingOption(self.msg).value

    @healing_option.setter
    def healing_option(self, healing_option):
        self.__stub.SetHealingOption(messages.int_property_message(self, healing_option))

    @property
    def defeature_absolute_length(self):
        """:obj:`str`: Absolute length used as tolerance when defeaturing polygons."""
        return self.__stub.GetDefeatureAbsoluteLength(self.msg).value

    @defeature_absolute_length.setter
    def defeature_absolute_length(self, defeature_absolute_length):
        self.__stub.SetDefeatureAbsoluteLength(
            messages.string_property_message(self, defeature_absolute_length)
        )

    @property
    def defeature_ratio(self):
        """:obj:`float`: Extent ratio used as tolerance when defeaturing polygons."""
        return self.__stub.GetDefeatureRatio(self.msg).value

    @defeature_ratio.setter
    def defeature_ratio(self, defeature_ratio):
        self.__stub.SetDefeatureRatio(messages.double_property_message(self, defeature_ratio))

    @property
    def small_void_area(self):
        """:obj:`float`: Voids with an area smaller than this value are ignored during simulation."""
        return self.__stub.GetSmallVoidArea(self.msg).value

    @small_void_area.setter
    def small_void_area(self, small_void_area):
        self.__stub.SetSmallVoidArea(messages.double_property_message(self, small_void_area))

    @property
    def via_model_type(self):
        """:class:`.ViaStyle`: Via model type."""
        return ViaStyle(self.__stub.GetViaModelType(self.msg).via_model_type)

    @via_model_type.setter
    def via_model_type(self, via_model_type):
        self.__stub.SetViaModelType(
            pb.ViaModelTypePropertyMessage(target=self.msg, via_model_type=via_model_type.value)
        )

    @property
    def num_via_sides(self):
        """:obj:`int`: Number of sides a via is considered to have."""
        return self.__stub.GetNumViaSides(self.msg).value

    @num_via_sides.setter
    def num_via_sides(self, num_via_sides):
        self.__stub.SetNumViaSides(messages.uint64_property_message(self, num_via_sides))

    @property
    def num_via_density(self):
        """:obj:`float`: Spacing between vias."""
        return self.__stub.GetViaDensity(self.msg).value

    @num_via_density.setter
    def num_via_density(self, num_via_density):
        self.__stub.SetViaDensity(messages.double_property_message(self, num_via_density))

    @property
    def via_material(self):
        """:obj:`str`: Default via material."""
        return self.__stub.GetViaMaterial(self.msg).value

    @via_material.setter
    def via_material(self, via_material):
        self.__stub.SetViaMaterial(messages.string_property_message(self, via_material))

    @property
    def mesh_for_via_plating(self):
        """:obj:`bool`: Flag indicating if meshing for via plating is enabled."""
        return self.__stub.GetMeshForViaPlating(self.msg).value

    @mesh_for_via_plating.setter
    def mesh_for_via_plating(self, mesh_for_via_plating):
        self.__stub.SetMeshForViaPlating(messages.bool_property_message(self, mesh_for_via_plating))

    @property
    def model_type(self):
        """:class:`.ModelType`: model type."""
        return ModelType(self.__stub.GetModelType(self.msg).defeature_model_type)

    @model_type.setter
    def model_type(self, model_type):
        self.__stub.SetModelType(
            pb.DefeatureModelTypePropertyMessage(
                target=self.msg, defeature_model_type=model_type.value
            )
        )


class AdvancedMeshingSettings(SimulationSettingsBase):
    """Class representing base advanced meshing simulation settings."""

    __stub: AdvancedMeshingSettingsServiceStub = StubAccessor(StubType.advanced_mesh_sim_settings)

    @property
    def arc_step_size(self):
        """:obj:`str`: Arc step size used when approximating arcs."""
        return self.__stub.GetArcStepSize(self.msg).value

    @arc_step_size.setter
    def arc_step_size(self, arc_step_size):
        self.__stub.SetArcStepSize(messages.string_property_message(self, arc_step_size))

    @property
    def circle_start_azimuth(self):
        """:obj:`str`: Starting azimuth used when approximating arcs."""
        return self.__stub.GetCircleStartAzimuth(self.msg).value

    @circle_start_azimuth.setter
    def circle_start_azimuth(self, circle_start_azimuth):
        self.__stub.SetCircleStartAzimuth(
            messages.string_property_message(self, circle_start_azimuth)
        )

    @property
    def max_num_arc_points(self):
        """:obj:`str`: Maximum number of points used to approximate arcs."""
        return self.__stub.GetMaxNumArcPoints(self.msg).value

    @max_num_arc_points.setter
    def max_num_arc_points(self, max_num_arc_points):
        self.__stub.SetMaxNumArcPoints(messages.uint64_property_message(self, max_num_arc_points))

    @property
    def use_arc_chord_error_approx(self):
        """:obj:`bool`: Flag indicating if arc chord error approximation is used."""
        return self.__stub.GetUseArcChordErrorApprox(self.msg).value

    @use_arc_chord_error_approx.setter
    def use_arc_chord_error_approx(self, use_arc_chord_error_approx):
        self.__stub.SetUseArcChordErrorApprox(
            messages.bool_property_message(self, use_arc_chord_error_approx)
        )

    @property
    def arc_to_chord_error(self):
        """:obj:`str`: Maximum allowable arc to chord error."""
        return self.__stub.GetArcChordErrorApprox(self.msg).value

    @arc_to_chord_error.setter
    def arc_to_chord_error(self, arc_to_chord_error):
        self.__stub.SetArcChordErrorApprox(
            messages.string_property_message(self, arc_to_chord_error)
        )


class SolverSettings(SimulationSettingsBase):
    """Class representing base solver simulation settings."""

    __stub: SolverSettingsServiceStub = StubAccessor(StubType.solver_sim_settings)

    @property
    def thin_signal_layer_threshold(self):
        """:obj:`str`: Value below which signal layers are considered to have zero thickness."""
        return self.__stub.GetThinSignalLayerThreshold(self.msg).value

    @thin_signal_layer_threshold.setter
    def thin_signal_layer_threshold(self, thin_signal_layer_threshold):
        self.__stub.SetThinSignalLayerThreshold(
            messages.string_property_message(self, thin_signal_layer_threshold)
        )

    @property
    def thin_dielectric_layer_threshold(self):
        """:obj:`str`: Value below which dielectric layers are merged with adjacent dielectric layers."""
        return self.__stub.GetThinDielectricLayerThreshold(self.msg).value

    @thin_dielectric_layer_threshold.setter
    def thin_dielectric_layer_threshold(self, thin_dielectric_layer_threshold):
        self.__stub.SetThinDielectricLayerThreshold(
            messages.string_property_message(self, thin_dielectric_layer_threshold)
        )



"""IC component property."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from src.ansys.edb.core.definition.die_property import DieProperty
    from src.ansys.edb.core.definition.solder_ball_property import SolderBallProperty

from ansys.api.edb.v1.ic_component_property_pb2_grpc import ICComponentPropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.definition import component_property, die_property, solder_ball_property
from ansys.edb.core.definition.port_property import PortProperty
from ansys.edb.core.inner import messages
from ansys.edb.core.session import StubAccessor, StubType


class ICComponentProperty(component_property.ComponentProperty):
    """Represents the properties of an IC component."""

    __stub: ICComponentPropertyServiceStub = StubAccessor(StubType.ic_component_property)

    @classmethod
    def create(cls) -> ICComponentProperty:
        """Create an IC component property.

        Returns
        -------
        .ICComponentProperty
        """
        return ICComponentProperty(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def solder_ball_property(self) -> SolderBallProperty:
        """:class:`.SolderBallProperty`: Solder ball properties of the IC component.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return solder_ball_property.SolderBallProperty(
            self.__stub.GetSolderBallProperty(messages.edb_obj_message(self))
        )

    @solder_ball_property.setter
    def solder_ball_property(self, value: SolderBallProperty):
        self.__stub.SetSolderBallProperty(
            messages.pointer_property_message(target=self, value=value)
        )

    @property
    def die_property(self) -> DieProperty:
        """:class:`.DieProperty`: Die properties of the IC component.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return die_property.DieProperty(self.__stub.GetDieProperty(messages.edb_obj_message(self)))

    @die_property.setter
    def die_property(self, value: DieProperty):
        self.__stub.SetDieProperty(messages.pointer_property_message(target=self, value=value))

    @property
    def port_property(self) -> PortProperty:
        """:class:`.PortProperty`: Port properties of the IC component.

        A copy is returned. Use the setter for any modifications to be reflected.
        """
        return PortProperty(self.__stub.GetPortProperty(messages.edb_obj_message(self)))

    @port_property.setter
    def port_property(self, value: PortProperty):
        self.__stub.SetPortProperty(messages.pointer_property_message(target=self, value=value))



"""This module allows for the creating of objects while avoid circular imports."""

from ansys.edb.core.edb_defs import LayoutObjType

_type_creator_params_dict = None
_primitive_type_creator_params_dict = None
_terminal_type_creator_params_dict = None


class _CreatorParams:
    def __init__(self, obj_type, do_cast=False):
        self.obj_type = obj_type
        self.do_cast = do_cast


def _initialize_type_creator_params_dict():
    global _type_creator_params_dict

    from ansys.edb.core.hierarchy.cell_instance import CellInstance
    from ansys.edb.core.hierarchy.group import Group
    from ansys.edb.core.hierarchy.pin_group import PinGroup
    from ansys.edb.core.inner.conn_obj import ConnObj
    from ansys.edb.core.layout.voltage_regulator import VoltageRegulator
    from ansys.edb.core.net.differential_pair import DifferentialPair
    from ansys.edb.core.net.extended_net import ExtendedNet
    from ansys.edb.core.net.net import Net
    from ansys.edb.core.net.net_class import NetClass
    from ansys.edb.core.primitive.padstack_instance import PadstackInstance
    from ansys.edb.core.primitive.primitive import Primitive
    from ansys.edb.core.terminal.terminal import Terminal
    from ansys.edb.core.terminal.terminal_instance import TerminalInstance

    _type_creator_params_dict = {
        LayoutObjType.PRIMITIVE: _CreatorParams(Primitive, True),
        LayoutObjType.PADSTACK_INSTANCE: _CreatorParams(PadstackInstance),
        LayoutObjType.TERMINAL: _CreatorParams(Terminal, True),
        LayoutObjType.TERMINAL_INSTANCE: _CreatorParams(TerminalInstance),
        LayoutObjType.CELL_INSTANCE: _CreatorParams(CellInstance),
        LayoutObjType.GROUP: _CreatorParams(Group, True),
        LayoutObjType.PIN_GROUP: _CreatorParams(PinGroup),
        LayoutObjType.VOLTAGE_REGULATOR: _CreatorParams(VoltageRegulator),
        LayoutObjType.NET_CLASS: _CreatorParams(NetClass),
        LayoutObjType.EXTENDED_NET: _CreatorParams(ExtendedNet),
        LayoutObjType.DIFFERENTIAL_PAIR: _CreatorParams(DifferentialPair),
        LayoutObjType.NET: _CreatorParams(Net),
        LayoutObjType.INVALID_LAYOUT_OBJ: _CreatorParams(ConnObj, True),
    }
    return _type_creator_params_dict


def _initialize_primitive_type_creator_params_dict():
    global _primitive_type_creator_params_dict

    from ansys.edb.core.primitive.board_bend_def import BoardBendDef
    from ansys.edb.core.primitive.bondwire import Bondwire
    from ansys.edb.core.primitive.circle import Circle
    from ansys.edb.core.primitive.path import Path
    from ansys.edb.core.primitive.polygon import Polygon
    from ansys.edb.core.primitive.primitive import PrimitiveType
    from ansys.edb.core.primitive.primitive_instance_collection import PrimitiveInstanceCollection
    from ansys.edb.core.primitive.rectangle import Rectangle
    from ansys.edb.core.primitive.text import Text

    _primitive_type_creator_params_dict = {
        PrimitiveType.BOARD_BEND: _CreatorParams(BoardBendDef),
        PrimitiveType.BONDWIRE: _CreatorParams(Bondwire),
        PrimitiveType.CIRCLE: _CreatorParams(Circle),
        PrimitiveType.PATH: _CreatorParams(Path),
        PrimitiveType.POLYGON: _CreatorParams(Polygon),
        PrimitiveType.RECTANGLE: _CreatorParams(Rectangle),
        PrimitiveType.TEXT: _CreatorParams(Text),
        PrimitiveType.PRIM_INST_COLLECTION: _CreatorParams(PrimitiveInstanceCollection),
    }
    return _primitive_type_creator_params_dict


def _initialize_terminal_type_creator_params_dict():
    global _terminal_type_creator_params_dict

    from ansys.edb.core.terminal.bundle_terminal import BundleTerminal
    from ansys.edb.core.terminal.edge_terminal import EdgeTerminal
    from ansys.edb.core.terminal.padstack_instance_terminal import PadstackInstanceTerminal
    from ansys.edb.core.terminal.pin_group_terminal import PinGroupTerminal
    from ansys.edb.core.terminal.point_terminal import PointTerminal
    from ansys.edb.core.terminal.terminal import TerminalType
    from ansys.edb.core.terminal.terminal_instance_terminal import TerminalInstanceTerminal

    _terminal_type_creator_params_dict = {
        TerminalType.BUNDLE: _CreatorParams(BundleTerminal),
        TerminalType.EDGE: _CreatorParams(EdgeTerminal),
        TerminalType.PADSTACK_INST: _CreatorParams(PadstackInstanceTerminal),
        TerminalType.PIN_GROUP: _CreatorParams(PinGroupTerminal),
        TerminalType.POINT: _CreatorParams(PointTerminal),
        TerminalType.TERM_INST: _CreatorParams(TerminalInstanceTerminal),
    }
    return _terminal_type_creator_params_dict


def _initialize_and_get_creator_dict(initializer, creator_dict):
    return initializer() if creator_dict is None else creator_dict


def _get_type_creator_dict():
    return _initialize_and_get_creator_dict(
        _initialize_type_creator_params_dict, _type_creator_params_dict
    )


def _get_primitive_type_creator_dict():
    return _initialize_and_get_creator_dict(
        _initialize_primitive_type_creator_params_dict, _primitive_type_creator_params_dict
    )


def _get_terminal_type_creator_dict():
    return _initialize_and_get_creator_dict(
        _initialize_terminal_type_creator_params_dict, _terminal_type_creator_params_dict
    )


def create_obj(msg, obj_type, do_cast):
    """Create an object from the provided message of the provided type."""
    obj = obj_type(msg)
    if do_cast:
        obj = obj.cast()
    return obj


def create_obj_from_creator_dict(creator_dict, msg, obj_type):
    """Create an object from the provided message of the type corresponding to the provided object type."""
    params = creator_dict[obj_type]
    return create_obj(msg, params.obj_type, params.do_cast)


def create_lyt_obj(msg, lyt_obj_type):
    """Create a layout object from the provided message of the type corresponding to the provided layout object type."""
    return create_obj_from_creator_dict(_get_type_creator_dict(), msg, lyt_obj_type)


def create_primitive(msg, prim_type):
    """Create a primitive from the provided message of the type corresponding to the provided primitive type."""
    return create_obj_from_creator_dict(_get_primitive_type_creator_dict(), msg, prim_type)


def create_terminal(msg, term_type):
    """Create a terminal from the provided message of the type corresponding to the provided terminal type."""
    return create_obj_from_creator_dict(_get_terminal_type_creator_dict(), msg, term_type)


def create_conn_obj(msg):
    """Create a connection object of its derived type based on its layout object type.

    Parameters
    ----------
    msg : ansys.api.edb.v1.edb_messages_pb2.EDBObjMessage

    Returns
    -------
    ansys.edb.core.inner.ConnObj
    """
    return create_lyt_obj(msg, LayoutObjType.INVALID_LAYOUT_OBJ)



"""Polygon data."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.typing import PointLike
    from ansys.edb.core.utility.value import Value
    from ansys.edb.core.geometry.point_data import PointData

from enum import Enum
import itertools
import math

from ansys.api.edb.v1 import edb_defs_pb2, point_data_pb2, polygon_data_pb2, polygon_data_pb2_grpc

from ansys.edb.core import session
from ansys.edb.core.geometry.arc_data import ArcData
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.inner.utils import client_stream_iterator
from ansys.edb.core.utility import conversions


class PolygonSenseType(Enum):
    """Provides an enum representing the direction of polygon sense."""

    SENSE_UNKNOWN = point_data_pb2.SENSE_UNKNOWN
    SENSE_CW = point_data_pb2.SENSE_CW
    SENSE_CCW = point_data_pb2.SENSE_CCW


class ExtentType(Enum):
    """Provides an enum representing extent types for geometries."""

    CONFORMING = edb_defs_pb2.CONFORMING
    BOUNDING_BOX = edb_defs_pb2.BOUNDING_BOX


class IntersectionType(Enum):
    """Provides an enum representing intersection types for geometries."""

    NO_INTERSECTION = polygon_data_pb2.NO_INTERSECTION
    THIS_INSIDE_OTHER = polygon_data_pb2.THIS_INSIDE_OTHER
    OTHER_INSIDE_THIS = polygon_data_pb2.OTHER_INSIDE_THIS
    COMMON_INTERSECTION = polygon_data_pb2.COMMON_INTERSECTION
    UNDEFINED_INTERSECTION = polygon_data_pb2.UNDEFINED_INTERSECTION


class PolygonData:
    """Represents a polygon data object."""

    @staticmethod
    def _polygon_data_request_iterator(polys):
        chunk_entry_creator = lambda poly: messages.polygon_data_message(poly)
        chunk_entries_getter = lambda chunk: chunk.polygons
        return client_stream_iterator(
            polys,
            polygon_data_pb2.PolygonDataListMessage,
            chunk_entry_creator,
            chunk_entries_getter,
        )

    __stub: polygon_data_pb2_grpc.PolygonDataServiceStub = session.StubAccessor(
        session.StubType.polygon_data
    )

    def __init__(
        self,
        points: PointLike = None,
        arcs: ArcData = None,
        lower_left: PointLike = None,
        upper_right: PointLike = None,
        holes: list[PolygonData] = None,
        sense: PolygonSenseType = PolygonSenseType.SENSE_CCW,
        closed: bool = True,
    ):
        """Create a polygon.

        Parameters
        ----------
        points : list of :term:`Point2DLike`, default: None
        arcs : list of .ArcData, default: None
        lower_left : :term:`Point2DLike`, default: None
        upper_right : :term:`Point2DLike`, default: None
        holes : list of .PolygonData, default: None
        sense : .PolygonSenseType, default: SENSE_CCW
        closed : bool, default: True
        """
        self._holes, self._sense, self._is_closed = (
            [] if holes is None else holes,
            PolygonSenseType(sense),
            closed,
        )

        if points is not None:
            self._points = [conversions.to_point(pt) for pt in points]
        elif arcs is not None:
            self._points = list(
                itertools.chain.from_iterable(
                    [[arc.start, arc.end] if arc.is_segment() else arc.points for arc in arcs]
                )
            )
        elif lower_left is not None and upper_right is not None:
            ll, ur = [conversions.to_point(pt) for pt in [lower_left, upper_right]]
            self._points = [
                conversions.to_point(pt)
                for pt in [(ll.x, ll.y), (ur.x, ll.y), (ur.x, ur.y), (ll.x, ur.y)]
            ]
        else:
            raise TypeError("PolygonData must be initialized from a list of points/arcs or a box.")

    def __len__(self) -> int:
        """Get the number of coordinates.

        Returns
        -------
        int
        """
        return len(self.points)

    @property
    def points(self) -> list[PointData]:
        """
        :obj:`list` of :class:`.PointData`: List of coordinates for the points.

        This property is read-only.
        """
        return self._points

    @property
    def holes(self) -> list[PolygonData]:
        """
        :obj:`list` of :class:`.PolygonData`: List of holes.

        This property is read-only.
        """
        return self._holes

    @property
    def arc_data(self) -> list[ArcData]:
        """
        :obj:`list` of :class:`.ArcData`: List of segments that represent the arc data of a polygon.

        This property is read-only.
        """
        i, n = 0, len(self)
        i_max = n if self.is_closed else n - 1
        segments = []

        while i < i_max:
            h, incr = 0, 1
            p1, p2 = self.points[i], self.points[(i + incr) % n]
            if p2.is_arc:
                h, incr = p2.arc_height.double, 2
                p2 = self.points[(i + incr) % n]
            segments.append(ArcData(p1, p2, height=h))
            i += incr
        return segments

    def _is_ccw(self) -> bool:
        return self._sense == PolygonSenseType.SENSE_CCW

    @property
    def is_closed(self) -> bool:
        """
        :obj:`bool`: Flag indicating if a polygon is closed between the first and last points.

        This property is read-only.
        """
        return self._is_closed

    @property
    def sense(self) -> PolygonSenseType:
        """
        :class:`PolygonSenseType`: Polygon sense type.

        This property is read-only.
        """
        return self._sense

    def is_hole(self) -> bool:
        """Determine whether the polygon is a hole.

        Returns
        -------
        bool
            ``True`` when the polygon is a hole, ``False`` otherwise.
        """
        return self.is_closed and not self._is_ccw()

    def is_parametric(self) -> bool:
        """Determine whether a polygon contains any parametrized points.

        Returns
        -------
        bool
            ``True`` when the polygon contains parametrized points, ``False`` otherwise.
        """
        return any(pt.is_parametric for pt in self.points)

    def has_arcs(self) -> bool:
        """Determine whether the polygon contains any arcs.

        Returns
        -------
        bool
            ``True`` when the polygon contains arcs, ``False`` otherwise.
        """
        return any(pt.is_arc for pt in self.points)

    def has_holes(self) -> bool:
        """Determine whether the polygon contains any holes.

        Returns
        -------
        bool
            ``True`` when the polygon contains holes, ``False`` otherwise.
        """
        return len(self.holes) > 0

    def is_circle(self) -> bool:
        """Determine whether the outer contour of the polygon is a circle.

        Returns
        -------
        bool
            ``True`` when the outer contour of the polygon is a circle holes, ``False`` otherwise.
        """
        return self.__stub.IsCircle(messages.polygon_data_message(self)).value

    def is_box(self) -> bool:
        """Determine whether the outer contour of the polygon is a box.

        Returns
        -------
        bool
            ``True`` when the outer corner of the polygon is a box, ``False`` otherwise.
        """
        return self.__stub.IsBox(messages.polygon_data_message(self)).value

    def is_convex(self) -> bool:
        """Determine whether the polygon is a convex hull.

        Returns
        -------
        bool
            ``True`` when the polygon is a convex hull, ``False`` otherwise.
        """
        return self.__stub.IsConvex(messages.polygon_data_message(self)).value

    def area(self) -> float:
        """Compute the area of the polygon.

        Returns
        -------
        float
            Area of the polygon.
        """
        return self.__stub.GetArea(messages.polygon_data_message(self)).value

    def has_self_intersections(self, tol: float = 1e-9) -> bool:
        """Determine whether the polygon contains any self-intersections.

        Parameters
        ----------
        tol : float, default: 1e-9
            Tolerance.

        Returns
        -------
        bool
            ``True`` when the polygon contains self-intersections, ``False`` otherwise.
        """
        return self.__stub.HasSelfIntersections(
            messages.polygon_data_with_tol_message(self, tol)
        ).value

    @parser.to_polygon_data_list
    def remove_self_intersections(self, tol: float = 1e-9) -> list[PolygonData]:
        """Remove self-intersections from this polygon.

        Parameters
        ----------
        tol : float, default: 1e-9
            Tolerance.

        Returns
        -------
        list of .PolygonData
            A list of non self-intersecting polygons.
        """
        return self.__stub.RemoveSelfIntersections(
            messages.polygon_data_with_tol_message(self, tol)
        )

    @parser.to_point_data_list
    def normalized(self) -> list[PointData]:
        """Get the normalized points of the polygon.

        Returns
        -------
        list of .PointData
        """
        return self.__stub.GetNormalizedPoints(messages.polygon_data_message(self))

    @parser.to_polygon_data
    def move(self, vector: PointLike) -> PolygonData:
        """Move the polygon by a vector.

        Parameters
        ----------
        vector : :term:`Point2DLike`
            Vector in the form: ``(x, y)``.

        Returns
        -------
        .PolygonData
        """
        return self.__stub.Transform(messages.polygon_data_transform_message("move", self, vector))

    @parser.to_polygon_data
    def rotate(self, angle: float, center: PointLike) -> PolygonData:
        """Rotate the polygon at a center by an angle.

        Parameters
        ----------
        angle : float
            Angle in radians.
        center : :term:`Point2DLike`
            Center.

        Returns
        -------
        .PolygonData
        """
        return self.__stub.Transform(
            messages.polygon_data_transform_message("rotate", self, angle, center)
        )

    @parser.to_polygon_data
    def scale(self, factor: float, center: PointLike) -> PolygonData:
        """Scale the polygon by a linear factor from a center.

        Parameters
        ----------
        factor : float
            Linear factor.
        center : :term:`Point2DLike`
            Center.

        Returns
        -------
        .PolygonData
        """
        return self.__stub.Transform(
            messages.polygon_data_transform_message("scale", self, factor, center)
        )

    @parser.to_polygon_data
    def mirror_x(self, x: float) -> PolygonData:
        """Mirror a polygon by x line.

        Parameters
        ----------
        x : float
            X line.

        Returns
        -------
        .PolygonData
        """
        return self.__stub.Transform(messages.polygon_data_transform_message("mirror_x", x))

    @parser.to_box
    def bbox(self) -> tuple[PointData, PointData]:
        """Compute the bounding box.

         Returns
         -------
        tuple of (.PointData, .PointData)
        """
        return self.__stub.GetBBox(messages.polygon_data_list_message([self]))

    @classmethod
    @parser.to_box
    def bbox_of_polygons(cls, polygons: list[PolygonData]) -> tuple[PointData, PointData]:
        """Compute the bounding box of a list of polygons.

        Parameters
        ----------
        polygons : list of .PolygonData
            List of polygons.

        Returns
        -------
        tuple of (.PointData, .PointData)
        """
        return cls.__stub.GetStreamedBBox(PolygonData._polygon_data_request_iterator(polygons))

    @parser.to_circle
    def bounding_circle(self) -> tuple[PointData, Value]:
        """Compute the bounding circle of the polygon.

        Returns
        -------
        tuple of (.PointData, .Value)
        """
        return self.__stub.GetBoundingCircle(messages.polygon_data_message(self))

    @classmethod
    @parser.to_polygon_data
    def convex_hull(cls, polygons: list[PolygonData]) -> PolygonData:
        """Compute the convex hull of the union of a list of polygons.

        Parameters
        ----------
        others : list of .PolygonData
            List of polygons.

        Returns
        -------
        .PolygonData
        """
        return cls.__stub.GetConvexHull(messages.polygon_data_list_message(polygons))

    @parser.to_polygon_data
    def without_arcs(
        self, max_chord_error: float = 0, max_arc_angle: float = math.pi / 6, max_points: int = 8
    ) -> PolygonData:
        """Get polygon data with all arcs removed.

        Parameters
        ----------
        max_chord_error : float, default: 0
        max_arc_angle : float, default: math.pi
        max_points : int, default: 8

        Returns
        -------
        .PolygonData
        """
        return self.__stub.RemoveArcs(
            messages.polygon_data_remove_arc_message(
                self, max_chord_error, max_arc_angle, max_points
            )
        )

    @parser.to_polygon_data
    def defeature(self, tol: float = 1e-9) -> PolygonData:
        """Defeature a polygon.

        Parameters
        ----------
        tol : float, default: 1e-9
            Tolerance.

        Returns
        -------
        .PolygonData
        """
        return self.__stub.Defeature(messages.polygon_data_with_tol_message(self, tol))

    def is_inside(self, point: PointLike) -> bool:
        """Determine whether the point is inside the polygon.

        Parameters
        ----------
        point : :term:`Point2DLike`

        Returns
        -------
        bool
            ``True`` if the point is inside the polygon, ``False`` otherwise.
        """
        return self.__stub.IsInside(messages.polygon_data_with_point_message(self, point)).value

    def intersection_type(self, other: PolygonData, tol: float = 1e-9) -> IntersectionType:
        """Get the intersection type with another polygon.

        Parameters
        ----------
        other : .PolygonData
            Other polygon.
        tol : float, default: 1e-9
            Tolerance.

        Returns
        -------
        .IntersectionType
        """
        return IntersectionType(
            self.__stub.GetIntersectionType(
                messages.polygon_data_pair_with_tolerance_message(self, other, tol)
            ).intersection_type
        )

    def circle_intersect(self, center: PointLike, radius: float) -> bool:
        """Determine whether the circle intersects with a polygon.

        Parameters
        ----------
        center : :term:`Point2DLike`
            Center.
        radius : float
            Radius.

        Returns
        -------
        bool
            ``True`` if the circle intersects with a polygon, ``False`` otherwise.
        """
        return self.__stub.CircleIntersectsPolygon(
            messages.polygon_data_with_circle_message(self, center, radius)
        ).value

    @parser.to_point_data
    def closest_point(self, point: PointLike) -> PointData:
        """Compute a point on the polygon that is closest to another point.

        Parameters
        ----------
        point : :term:`Point2DLike`
           Other point.

        Returns
        -------
        .PointData
            Point closest to the given point.
        """
        return self.__stub.GetClosestPoints(
            messages.polygon_data_with_points_message(self, point=point)
        ).points[0]

    @parser.to_point_data_list
    def closest_points(self, polygon: PolygonData) -> tuple[PointData, PointData]:
        """Compute points on this and another polygon that are closest to the other polygon.

        Parameters
        ----------
        polygon : .PolygonData

        Returns
        -------
        tuple of (.PointData, .PointData)
        """
        return self.__stub.GetClosestPoints(
            messages.polygon_data_with_points_message(self, polygon=polygon)
        ).points

    @classmethod
    @parser.to_polygon_data_list
    def unite(cls, polygons: list[PolygonData]) -> list[PolygonData]:
        """Compute the union of a list of polygons.

        Parameters
        ----------
        polygons : list of .PolygonData
            List of polygons.

        Returns
        -------
        list of .PolygonData
        """
        return cls.__stub.GetUnion(messages.polygon_data_list_message(polygons))

    @classmethod
    @parser.to_polygon_data_list
    def intersect(
        cls, polygons1: list[PolygonData] | PolygonData, polygons2: list[PolygonData] | PolygonData
    ) -> list[PolygonData]:
        """Compute the intersection of one or more lists of polygons.

        Parameters
        ----------
        polygons1 : list of .PolygonData or .PolygonData
            First list of polygons.
        polygons2 : list of .PolygonData or .PolygonData
            Second optional list of polygons.

        Returns
        -------
        list of .PolygonData
        """
        return cls.__stub.GetIntersection(messages.polygon_data_pair_message(polygons1, polygons2))

    @classmethod
    @parser.to_polygon_data_list
    def subtract(
        cls, polygons1: list[PolygonData] | PolygonData, polygons2: list[PolygonData] | PolygonData
    ) -> list[PolygonData]:
        """Subtract a set of polygons from another set of polygons.

        Parameters
        ----------
        polygons1 : list of .PolygonData or PolygonData
            List of base polygons.
        polygons2 : list of .PolygonData or PolygonData
            List of polygons to subtract.

        Returns
        -------
        list of .PolygonData
        """
        return cls.__stub.Subtract(messages.polygon_data_pair_message(polygons1, polygons2))

    @classmethod
    @parser.to_polygon_data_list
    def xor(
        cls, polygons1: list[PolygonData] | PolygonData, polygons2: list[PolygonData] | PolygonData
    ) -> list[PolygonData]:
        """Compute an exclusive OR between a set of polygons and another set of polygons.

        Parameters
        ----------
        polygons1 : list of .PolygonData or PolygonData
            First list of polygons.
        polygons2 : list of .PolygonData or PolygonData
            Second list of polygons.

        Returns
        -------
        list of .PolygonData
        """
        return cls.__stub.Xor(messages.polygon_data_pair_message(polygons1, polygons2))

    @parser.to_polygon_data_list
    def expand(
        self, offset: float, round_corner: bool, max_corner_ext: float, tol: float = 1e-9
    ) -> list[PolygonData]:
        """Expand the polygon by an offset.

        Parameters
        ----------
        offset : float
            Expansion offset. Specify a negative value to shrink the polygon.
        round_corner : bool
            Whether the corners are rounded corners. If ``False``, the corners
            are straight edges.
        max_corner_ext : float
            Maximum corner extension to clip the corner at.
        tol : float, default: 1e-9
            Tolerance.

        Returns
        -------
        list of .PolygonData
        """
        return self.__stub.Expand(
            messages.polygon_data_expand_message(self, offset, tol, round_corner, max_corner_ext)
        )

    @classmethod
    @parser.to_polygon_data_list
    def alpha_shape(cls, points: list[PointLike], alpha: float) -> list[PolygonData]:
        """Compute the outline of a 2D point cloud using alpha shapes.

        Parameters
        ----------
        points : list of :term:`Point2DLike`
            List of points.
        alpha : float

        Returns
        -------
        list of .PolygonData
        """
        return cls.__stub.Get2DAlphaShape(
            messages.polygon_data_get_alpha_shape_message(points, alpha)
        )



"""Primitive classes."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.net.net import Net
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.typing import ValueLike
    from ansys.edb.core.hierarchy.pin_group import PinGroup


from enum import Enum

from ansys.api.edb.v1 import padstack_instance_pb2, padstack_instance_pb2_grpc

from ansys.edb.core.definition.padstack_def import PadstackDef
from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import conn_obj, messages
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal import padstack_instance_terminal
from ansys.edb.core.utility.layer_map import LayerMap
from ansys.edb.core.utility.value import Value


class BackDrillType(Enum):
    """Provides an enum representing back drill types."""

    NO_DRILL = padstack_instance_pb2.NO_DRILL
    LAYER_DRILL = padstack_instance_pb2.LAYER_DRILL
    DEPTH_DRILL = padstack_instance_pb2.DEPTH_DRILL


class PadstackInstance(conn_obj.ConnObj):
    """Represents a padstack instance object."""

    __stub: padstack_instance_pb2_grpc.PadstackInstanceServiceStub = StubAccessor(
        StubType.padstack_instance
    )
    layout_obj_type = LayoutObjType.PADSTACK_INSTANCE
    """:class:`.LayoutObjType`: Layout object type of the PadstackInstance class."""

    @classmethod
    def create(
        cls,
        layout: Layout,
        net: Net,
        name: str,
        padstack_def: PadstackDef,
        position_x: ValueLike,
        position_y: ValueLike,
        rotation: ValueLike,
        top_layer: Layer,
        bottom_layer: Layer,
        solder_ball_layer: Layer | None = None,
        layer_map: LayerMap | None = None,
    ) -> PadstackInstance:
        """Create a padstack instance.

        Parameters
        ----------
        layout : .Layout
            Layout to create the padstack instance in.
        net : .Net
            Net of the padstack instance.
        name : str
            Name of the padstack instance.
        padstack_def : .PadstackDef
            Padstack definition of the padstack instance.
        position_x : :term:`ValueLike`
            Position x of the padstack instance.
        position_y : :term:`ValueLike`
            Position y of the padstack instance.
        rotation : :term:`ValueLike`
            Rotation of the padstack instance.
        top_layer : .Layer
            Top layer of the padstack instance.
        bottom_layer : .Layer
            Bottom layer of the padstack instance.
        solder_ball_layer : .Layer
            Solder ball layer of the padstack instance or ``None`` for none.
        layer_map : .LayerMap
            Layer map of the padstack instance. ``None`` or empty results in
            auto-mapping.

        Returns
        -------
        PadstackInstance
            Padstack instance created.
        """
        padstack_instance = PadstackInstance(
            cls.__stub.Create(
                padstack_instance_pb2.PadstackInstCreateMessage(
                    layout=layout.msg,
                    net=net.msg,
                    name=name,
                    padstack_def=padstack_def.msg,
                    rotation=messages.value_message(rotation),
                    top_layer=top_layer.msg,
                    bottom_layer=bottom_layer.msg,
                    solder_ball_layer=messages.edb_obj_message(solder_ball_layer),
                    layer_map=messages.edb_obj_message(layer_map),
                )
            )
        )
        padstack_instance.set_position_and_rotation(position_x, position_y, rotation)
        return padstack_instance

    @property
    def padstack_def(self) -> PadstackDef:
        """:class:`.PadstackDef`: \
        Definition of the padstack instance.

        This property is read-only.
        """
        return PadstackDef(self.__stub.GetPadstackDef(self.msg))

    @property
    def name(self) -> str:
        """:obj:`str`: Name of the padstack instance."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, name: str):
        self.__stub.SetName(messages.edb_obj_name_message(self, name))

    def get_position_and_rotation(self) -> tuple[Value, Value, Value]:
        """Get the position and rotation of the padstack instance.

        Returns
        -------
        tuple of (.Value, .Value, .Value)

            Returns a tuple in this format:

            **(x, y, rotation)**

            **x** : X coordinate.

            **y** : Y coordinate.

            **rotation** : Rotation in radians.
        """
        params = self.__stub.GetPositionAndRotation(self.msg)
        return (
            Value(params.position.x),
            Value(params.position.y),
            Value(params.rotation),
        )

    def set_position_and_rotation(self, x: ValueLike, y: ValueLike, rotation: ValueLike):
        """Set the position and rotation of the padstack instance.

        Parameters
        ----------
        x : :term:`ValueLike`
            x : X coordinate.
        y : :term:`ValueLike`
            y : Y coordinate.
        rotation : :term:`ValueLike`
            Rotation in radians.
        """
        self.__stub.SetPositionAndRotation(
            padstack_instance_pb2.PadstackInstSetPositionAndRotationMessage(
                target=self.msg,
                params=padstack_instance_pb2.PadstackInstPositionAndRotationMessage(
                    position=messages.point_message((x, y)),
                    rotation=messages.value_message(rotation),
                ),
            )
        )

    def get_layer_range(self) -> tuple[Layer, Layer]:
        """Get the top and bottom layers of the padstack instance.

        Returns
        -------
        tuple of (.Layer, .Layer)
            The tuple is in this format:

            **(top_layer, bottom_layer)**

            **top_layer**: Top layer of the padstack instance
            **bottom_layer**: Bottom layer of the padstack instance
        """
        params = self.__stub.GetLayerRange(self.msg)
        return (
            Layer(params.top_layer).cast(),
            Layer(params.bottom_layer).cast(),
        )

    def set_layer_range(self, top_layer: Layer, bottom_layer: Layer):
        """Set the top and bottom layers of the padstack instance.

        Parameters
        ----------
        top_layer : .Layer
            Top layer of the padstack instance.
        bottom_layer : .Layer
            Bottom layer of the padstack instance.
        """
        self.__stub.SetLayerRange(
            padstack_instance_pb2.PadstackInstSetLayerRangeMessage(
                target=self.msg,
                range=padstack_instance_pb2.PadstackInstLayerRangeMessage(
                    top_layer=top_layer.msg,
                    bottom_layer=bottom_layer.msg,
                ),
            )
        )

    @property
    def solderball_layer(self) -> Layer:
        """:class:`.Layer`: Solderball layer of the padstack instance."""
        sb_layer = Layer(self.__stub.GetSolderBallLayer(self.msg))
        return sb_layer if sb_layer.is_null() else sb_layer.cast()

    @solderball_layer.setter
    def solderball_layer(self, solderball_layer: Layer):
        self.__stub.SetSolderBallLayer(
            padstack_instance_pb2.PadstackInstSetSolderBallLayerMessage(
                target=self.msg,
                layer=solderball_layer.msg,
            )
        )

    @property
    def layer_map(self) -> LayerMap:
        """:class:`.LayerMap`: Layer map of the padstack instance."""
        return LayerMap(self.__stub.GetLayerMap(self.msg))

    @layer_map.setter
    def layer_map(self, layer_map: LayerMap):
        self.__stub.SetLayerMap(messages.pointer_property_message(self, layer_map))

    def get_hole_overrides(self) -> tuple[bool, Value]:
        """Get the hole overrides of the padstack instance.

        Returns
        -------
        tuple of (bool, .Value)

            Returns a tuple in this format:

            **(is_hole_override, hole_override)**

            **is_hole_override** : If padstack instance is hole override.

            **hole_override** : Hole override diameter of this padstack instance.
        """
        params = self.__stub.GetHoleOverrides(self.msg)
        return (
            params.is_hole_override,
            Value(params.hole_override),
        )

    def set_hole_overrides(self, is_hole_override: bool, hole_override: ValueLike):
        """Set the hole overrides of the padstack instance.

        Parameters
        ----------
        is_hole_override : bool
            Whether the padstack instance is a hole override.
        hole_override : :term:`ValueLike`
            Hole override diameter of the padstack instance.
        """
        self.__stub.SetHoleOverrides(
            padstack_instance_pb2.PadstackInstSetHoleOverridesMessage(
                target=self.msg,
                hole_override_msg=padstack_instance_pb2.PadstackInstHoleOverridesMessage(
                    is_hole_override=is_hole_override,
                    hole_override=messages.value_message(hole_override),
                ),
            )
        )

    @property
    def is_layout_pin(self):
        """:obj:`bool`: Flag indicating if the padstack instance is a layout pin."""
        return self.__stub.GetIsLayoutPin(self.msg).value

    @is_layout_pin.setter
    def is_layout_pin(self, is_layout_pin):
        self.__stub.SetIsLayoutPin(
            padstack_instance_pb2.PadstackInstSetIsLayoutPinMessage(
                target=self.msg,
                is_layout_pin=is_layout_pin,
            )
        )

    def get_back_drill_type(self, from_bottom: bool) -> BackDrillType:
        """Get the back drill type of the padstack instance.

        Parameters
        ----------
        from_bottom : bool
            Whether to get the back drill type from the bottom.

        Returns
        -------
        .BackDrillType
            Back drill type of the padastack instance.
        """
        return BackDrillType(
            self.__stub.GetBackDrillType(
                PadstackInstance._get_back_drill_message(self, from_bottom)
            ).type
        )

    def get_back_drill_by_layer(self, from_bottom: bool) -> tuple[Layer, Value, Value]:
        """Get the back drill type by the layer.

        Parameters
        ----------
        from_bottom : bool
            Whether to get the back drill type from the bottom.

        Returns
        -------
        tuple of (.Layer, .Value, .Value)

            Returns a tuple in this format:

            **(drill_to_layer, offset, diameter)**

            **drill_to_layer** : Layer drills to. If drill from top, drill stops at the upper elevation of the layer.\
            If from bottom, drill stops at the lower elevation of the layer.

            **offset** : Layer offset (or depth if layer is empty).

            **diameter** : Drilling diameter.
        """
        params = self.__stub.GetBackDrillByLayer(
            PadstackInstance._get_back_drill_message(self, from_bottom)
        )

        return (
            Layer(params.drill_to_layer).cast(),
            Value(params.offset),
            Value(params.diameter),
        )

    def set_back_drill_by_layer(
        self, drill_to_layer: Layer, offset: ValueLike, diameter: ValueLike, from_bottom: bool
    ):
        """Set the back drill by the layer.

        Parameters
        ----------
        drill_to_layer : .Layer
            Layer to drill to. If drilling from the top, the drill stops at the upper
            elevation of the layer. If drilling from the bottom, the drill stops at
            the lower elevation of the layer.
        offset : :term:`ValueLike`
            Layer offset (or depth if the layer is empty).
        diameter : :term:`ValueLike`
            Drilling diameter.
        from_bottom : bool
            Whether to set the back drill type from the bottom.
        """
        self.__stub.SetBackDrillByLayer(
            padstack_instance_pb2.PadstackInstSetBackDrillByLayerMessage(
                target=self.msg,
                drill_to_layer=drill_to_layer.msg,
                offset=messages.value_message(offset),
                diameter=messages.value_message(diameter),
                from_bottom=from_bottom,
            )
        )

    def get_back_drill_by_depth(self, from_bottom: bool) -> tuple[Value, Value]:
        """Get the back drill type by depth.

        Parameters
        ----------
        from_bottom : bool
            Whether to get the back drill type from the bottom.

        Returns
        -------
        tuple of (.Value, .Value)
            Returns a tuple in this format:

            **(drill_depth, diameter)**

            **drill_depth** : Drilling depth, may not align with layer.

            **diameter** : Drilling diameter.
        """
        params = self.__stub.GetBackDrillByDepth(
            PadstackInstance._get_back_drill_message(self, from_bottom)
        )
        return Value(params.drill_depth), Value(params.diameter)

    def set_back_drill_by_depth(
        self, drill_depth: ValueLike, diameter: ValueLike, from_bottom: bool
    ):
        """Set the back drill type by depth.

        Parameters
        ----------
        drill_depth : :term:`ValueLike`
            Drilling depth, which may not align with the layer.
        diameter : :term:`ValueLike`
            Drilling diameter.
        from_bottom : bool
            Whether to set the back drill type from the bottom.
        """
        self.__stub.SetBackDrillByDepth(
            padstack_instance_pb2.PadstackInstSetBackDrillByDepthMessage(
                target=self.msg,
                drill_depth=messages.value_message(drill_depth),
                diameter=messages.value_message(diameter),
                from_bottom=from_bottom,
            )
        )

    def get_padstack_instance_terminal(self) -> padstack_instance_terminal.PadstackInstanceTerminal:
        """:class:`.PadstackInstanceTerminal`: \
        Terminal of the padstack instance."""
        return padstack_instance_terminal.PadstackInstanceTerminal(
            self.__stub.GetPadstackInstanceTerminal(self.msg)
        )

    def is_in_pin_group(self, pin_group: PinGroup) -> bool:
        """Determine if the padstack instance is in a given pin group.

        Parameters
        ----------
        pin_group : .PinGroup
            Pin group to check if the padstack instance is in it.

        Returns
        -------
        bool
            Whether the padstack instance is in a pin group.
        """
        return self.__stub.IsInPinGroup(
            padstack_instance_pb2.PadstackInstIsInPinGroupMessage(
                target=self.msg,
                pin_group=pin_group.msg,
            )
        ).value

    @property
    def pin_groups(self) -> list[PinGroup]:
        """:obj:`list` of :class:`.PinGroup`: \
        Pin groups of the padstack instance.

        This property is read-only.
        """
        from ansys.edb.core.hierarchy import pin_group

        pins = self.__stub.GetPinGroups(self.msg).items
        return [pin_group.PinGroup(p) for p in pins]

    @staticmethod
    def _get_back_drill_message(padstack_inst: PadstackInstance, from_bottom: bool):
        return padstack_instance_pb2.PadstackInstGetBackDrillMessage(
            target=padstack_inst.msg,
            from_bottom=from_bottom,
        )



"""S-parameter model."""
from ansys.edb.core.hierarchy.model import Model
from ansys.edb.core.inner import messages
from ansys.edb.core.session import SParameterModelServiceStub, StubAccessor, StubType


class SParameterModel(Model):
    """Represents an S-parameter model object."""

    __stub: SParameterModelServiceStub = StubAccessor(StubType.sparameter_model)

    @classmethod
    def create(cls, name, ref_net):
        """Create an S-parameter model.

        Parameters
        ----------
        name : str
            Name of the S-parameter model.
        ref_net : str
            Name of the reference net.
        """
        return cls(cls.__stub.Create(messages.sparameter_model_message(name, ref_net)))

    @property
    def _properties(self):
        return self.__stub.GetProperties(messages.edb_obj_message(self))

    @property
    def component_model(self):
        """:obj:`str`: Name of the component model."""
        return self._properties.name

    @component_model.setter
    def component_model(self, name):
        self.__stub.SetComponentModelName(messages.string_property_message(self, name))

    @property
    def reference_net(self):
        """:obj:`str`: Name of the reference net."""
        return self._properties.ref_net

    @reference_net.setter
    def reference_net(self, name):
        self.__stub.SetReferenceNet(messages.string_property_message(self, name))



"""EDB Error Manager."""

from enum import Enum
from typing import List

import ansys.api.edb.v1.edb_error_manager_pb2 as pb2
from ansys.api.edb.v1.edb_error_manager_pb2_grpc import EDBErrorManagerServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.inner.utils import stream_items_from_server
from ansys.edb.core.session import StubAccessor, StubType

__stub_accessor = StubAccessor(StubType.edb_error_manager)


def _get_stub() -> EDBErrorManagerServiceStub:
    return __stub_accessor.__get__()


class EDBErrorSeverity(Enum):
    """Enum representing the severity of an :class:`.EDBError`."""

    INFO = pb2.EDBErrorSeverity.INFO
    WARNING = pb2.EDBErrorSeverity.WARNING
    ERROR = pb2.EDBErrorSeverity.ERROR
    FATAL = pb2.EDBErrorSeverity.FATAL


class EDBError:
    """Class representing an error received from the EDB error manager on the server."""

    def __init__(self, message: str, severity: EDBErrorSeverity):
        """Construct an EDBError object.

        Parameters
        ----------
        message : str
        severity : .EDBErrorSeverity
        """
        self._message = message
        self._severity = severity

    @property
    def message(self) -> str:
        """:obj:`str`: The message stored in the error."""
        return self._message

    @property
    def severity(self) -> EDBErrorSeverity:
        """:class:`.EDBErrorSeverity`: The severity of the error."""
        return self._severity


def _msg_to_edb_error(msg):
    return EDBError(msg.message, EDBErrorSeverity(msg.severity))


def get_error_messages() -> List[EDBError]:
    """Get all the error messages stored in the EDB error manager. All existing \
    messages in the EDB error manager are cleared.

    Returns
    -------
    list of .EDBError
    """
    return stream_items_from_server(
        _msg_to_edb_error, _get_stub().GetErrors(empty_pb2.Empty()), "messages"
    )



"""Common definitions used in EDB."""

from enum import Enum

import ansys.api.edb.v1.definition_obj_pb2 as definition_obj_pb2
import ansys.api.edb.v1.layout_obj_pb2 as layout_obj_pb2


class LayoutObjType(Enum):
    """Enum representing layout object types."""

    INVALID_LAYOUT_OBJ = layout_obj_pb2.INVALID_LAYOUT_OBJ
    PRIMITIVE = layout_obj_pb2.PRIMITIVE
    PADSTACK_INSTANCE = layout_obj_pb2.PADSTACK_INST
    TERMINAL = layout_obj_pb2.TERM
    TERMINAL_INSTANCE = layout_obj_pb2.TERM_INST
    CELL_INSTANCE = layout_obj_pb2.CELL_INST
    LAYER = layout_obj_pb2.LAYER
    NET = layout_obj_pb2.NET
    PADSTACK = layout_obj_pb2.PADSTACK
    GROUP = layout_obj_pb2.GROUP
    NET_CLASS = layout_obj_pb2.NET_CLASS
    CELL = layout_obj_pb2.CELL
    DIFFERENTIAL_PAIR = layout_obj_pb2.DIFFERENTIAL_PAIR
    PIN_GROUP = layout_obj_pb2.PIN_GROUP
    VOLTAGE_REGULATOR = layout_obj_pb2.VOLTAGE_REGULATOR
    EXTENDED_NET = layout_obj_pb2.EXTENDED_NET


class DefinitionObjType(Enum):
    """Enum representing definition object types."""

    INVALID_DEFINITION_TYPE = definition_obj_pb2.INVALID_DEFINITION_TYPE
    PADSTACK_DEF = definition_obj_pb2.PADSTACK_DEF
    COMPONENT_DEF = definition_obj_pb2.COMPONENT_DEF
    BONDWIRE_DEF = definition_obj_pb2.BONDWIRE_DEF
    MATERIAL_DEF = definition_obj_pb2.MATERIAL_DEF
    DATASET_DEF = definition_obj_pb2.DATASET_DEF
    PACKAGE_DEF = definition_obj_pb2.PACKAGE_DEF
    DEFINITION_TYPE_COUNT = definition_obj_pb2.DEFINITION_TYPE_COUNT



"""Component definition."""
from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from src.ansys.edb.core.database import ProductIdType

from ansys.api.edb.v1.component_def_pb2_grpc import ComponentDefServiceStub

from ansys.edb.core.definition import component_model, component_pin
from ansys.edb.core.edb_defs import DefinitionObjType
from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.inner.messages import (
    edb_obj_collection_property_message,
    edb_obj_pair_message,
    get_product_property_ids_message,
    get_product_property_message,
    set_product_property_message,
)
from ansys.edb.core.inner.utils import map_list
from ansys.edb.core.session import StubAccessor, StubType


class ComponentDef(ObjBase):
    """Represents a component definition."""

    __stub: ComponentDefServiceStub = StubAccessor(StubType.component_def)

    @classmethod
    def create(cls, db, comp_def_name, fp):
        """Create a component definition in a given database.

        Parameters
        ----------
        db : :class:`.Database`
            Database to create the component definition in.
        comp_def_name : str
            Name of the component definition to create.
        fp : :class:`.Cell`
            Footprint cell of the component definition, optional

        Returns
        -------
        ComponentDef
            Component definition created.
        """
        return ComponentDef(
            cls.__stub.Create(messages.component_def_creation_message(db, comp_def_name, fp))
        )

    @classmethod
    def find(cls, db, comp_def_name):
        """Find a component definition in a given database.

        Parameters
        ----------
        db : :class:`.Database`
            Database to search for the component definition.
        comp_def_name : str
            Name of the component definition.

        Returns
        -------
        ComponentDef
            Component definition found, ``None`` otherwise.
        """
        return ComponentDef(
            cls.__stub.FindByName(messages.object_name_in_layout_message(db, comp_def_name))
        )

    @property
    def definition_type(self):
        """:class:`DefinitionObjType`: Definition type."""
        return DefinitionObjType.COMPONENT_DEF

    @property
    def name(self):
        """:obj:`str`: Name of the component definition."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, value):
        self.__stub.SetName(messages.string_property_message(self, value))

    @property
    def footprint(self):
        """:class:`.Cell`: Footprint of the component definition."""
        from ansys.edb.core.layout import cell

        return cell.Cell(self.__stub.GetFootprintCell(self.msg))

    @footprint.setter
    def footprint(self, value):
        self.__stub.SetFootprintCell(messages.edb_obj_pair_message(self, value))

    @property
    def component_models(self):
        """:obj:`list` of :class:`.ComponentModel`: All component models associated with the component definition.

        This property is read-only.
        """
        objs = self.__stub.GetComponentModels(self.msg).items
        return map_list(objs, lambda msg: component_model.ComponentModel(msg).cast())

    @property
    def component_pins(self):
        """:obj:`list` of :class:`.ComponentPin`: All component pins of the component definition.

        This property is read-only.
        """
        objs = self.__stub.GetComponentPins(self.msg).items
        return map_list(objs, component_pin.ComponentPin)

    def add_component_model(self, value):
        """Add a component model to this component def.

        Parameters
        ----------
        value : :class:`Component Model <ansys.edb.core.definition.ComponentModel>`
            Component Model to be added.

        Notes
        -----
        Once a component model is added to one component def, it cannot be added to any other, even when removed.
        """
        self.__stub.AddComponentModel(messages.edb_obj_pair_message(self, value))

    def remove_component_model(self, value):
        """Remove a component model from this component def.

        Parameters
        ----------
        value : :class:`Component Model <ansys.edb.core.definition.ComponentModel>`
            Component Model to be removed.

        Notes
        -----
        Once a component model is added to one component def, it cannot be added to any other, even when removed.
        """
        self.__stub.RemoveComponentModel(messages.pointer_property_message(self, value))

    def reorder_pins(self, reordered_pins: List[component_pin.ComponentPin]):
        """Reorders the existing pins in the components definition to be in the same order \
        as in the provided list.

        Parameters
        ----------
        reordered_pins : list of .ComponentPin
            The component pins in the new order. These must be the same pins \
            that already exist in the component definition.
        """
        self.__stub.ReorderPins(edb_obj_collection_property_message(self, reordered_pins))

    def remove_pin(self, pin_to_remove: component_pin.ComponentPin):
        """Remove the provided component pin from the component definition. \
        the pin will be deleted and set to :meth:`null <.is_null>`.

        Parameters
        ----------
        pin_to_remove : .ComponentPin
            The pin to be removed.
        """
        self.__stub.RemovePin(edb_obj_pair_message(self, pin_to_remove))
        pin_to_remove.msg = None

    def get_product_property(self, prod_id: ProductIdType, attr_it: int) -> str:
        """Get the product property for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : .ProductIdType
            Product ID.
        attr_it : int
            Attribute ID.

        Returns
        -------
        str
            Product property for the given product ID and attribute ID.
        """
        return self.__stub.GetProductProperty(
            get_product_property_message(self, prod_id, attr_it)
        ).value

    def set_product_property(self, prod_id: ProductIdType, attr_it: int, prop_value: str):
        """Set the product property for the given product ID and attribute ID.

        Parameters
        ----------
        prod_id : .ProductIdType
            Product ID.
        attr_it : int
            Attribute ID.
        prop_value : str
            New value for the product property.
        """
        self.__stub.SetProductProperty(
            set_product_property_message(self, prod_id, attr_it, prop_value)
        )

    def get_product_property_ids(self, prod_id: ProductIdType) -> List[int]:
        """Get the list of property IDs for a given property ID.

        Parameters
        ----------
        prod_id : ProductIdType
            Product ID.

        Returns
        -------
        list[int]
            Attribute IDs for the given product ID.
        """
        attr_ids = self.__stub.GetProductPropertyIds(
            get_product_property_ids_message(self, prod_id)
        ).ids
        return [attr_id for attr_id in attr_ids]



"""This package defines exception base classes."""

from enum import Enum


class ErrorCode(Enum):
    """Provides EDB exception types."""

    UNKNOWN = "Unknown exception: {}."
    UNAVAILABLE = (
        "EDB server is not accessible. Make sure an instance is listening on the specified port.."
    )
    NO_SESSIONS = "No active session is detected."
    STARTUP_UNEXPECTED = "An unexpected error occurred when starting the local server: {}."
    STARTUP_TIMEOUT = "Could not start local server: Time out"
    STARTUP_MULTI_SESSIONS = "There can be only one session active at a time."
    STARTUP_NO_EXECUTABLE = (
        "Could not find necessary executables. Make sure the Ansys EM root directory is correct."
    )
    STARTUP_FAILURE_LICENSE = "Could not start the local server: No valid license detected."
    STARTUP_FAILURE_EDB = "Could not start the local server: Failed to initialize EDB."
    STARTUP_FAILURE = "Could not start the local server due to unknown reason."

    INVALID_ARGUMENT = "{}"


def _message(code, *args):
    template = code.value
    return template.format(*args)


class EDBSessionException(Exception):
    """Provides the base class for exceptions related to EDB sessions."""

    def __init__(self, code, *args):
        """Initialize EDBSessionException."""
        super().__init__(_message(code, *args))
        self._code = code


class InvalidArgumentException(EDBSessionException):
    """Provides the exception that occurs when a request fails due to an invalid argument."""

    def __init__(self, response):
        """Initialize InvalidArgumentException."""
        super().__init__(ErrorCode.INVALID_ARGUMENT, response.details())
        self._response = response



"""RaptorX simulation settings."""

import ansys.api.edb.v1.raptor_x_simulation_settings_pb2 as pb

from ansys.edb.core.inner import messages
from ansys.edb.core.session import (
    RaptorXAdvancedSettingsServiceStub,
    RaptorXGeneralSettingsServiceStub,
    StubAccessor,
    StubType,
)
from ansys.edb.core.simulation_setup.simulation_settings import (
    SimulationSettings,
    SimulationSettingsBase,
)


def _translate_options_dictionary(dictionary):
    return {
        key: messages.strings_message([val for val in vals]) for (key, vals) in dictionary.items()
    }


def _to_options_dict(msg):
    return {key: [val for val in vals.strings] for (key, vals) in msg.options.items()}


def _to_raptor_x_sim_settings_options_property_msg(obj, options):
    return pb.RaptorXSimSettingsOptionsPropertyMessage(
        target=obj.msg,
        value=pb.RaptorXSimSettingsOptionsMessage(
            options={
                key: messages.strings_message([val for val in vals])
                for (key, vals) in options.items()
            }
        ),
    )


class RaptorXSimulationSettings(SimulationSettings):
    """Represents SIWave simulation settings."""

    @property
    def general(self):
        """:class:`.RaptorXGeneralSettings`: General settings for RaptorX simulations."""
        return RaptorXGeneralSettings(self._sim_setup)

    @property
    def advanced(self):
        """:class:`.RaptorXAdvancedSettings`: Advanced settings for RaptorX simulations."""
        return RaptorXAdvancedSettings(self._sim_setup)


class RaptorXGeneralSettings(SimulationSettingsBase):
    """Represents general settings for RaptorX simulations."""

    __stub: RaptorXGeneralSettingsServiceStub = StubAccessor(StubType.raptor_x_general_sim_settings)

    @property
    def use_gold_em_solver(self):
        """:obj:`bool`: Flag indicating if the gold em solver is used. If ``False``, \
        the fast em solver is used."""
        return self.__stub.GetUseGoldEMSolver(self.msg).value

    @use_gold_em_solver.setter
    def use_gold_em_solver(self, use_gold_em_solver):
        self.__stub.SetUseGoldEMSolver(messages.bool_property_message(self, use_gold_em_solver))

    @property
    def max_frequency(self):
        """:obj:`str`: Maximum frequency value, which controls the tightness of the model mesh."""
        return self.__stub.GetMaxFrequency(self.msg).value

    @max_frequency.setter
    def max_frequency(self, max_frequency):
        self.__stub.SetMaxFrequency(messages.string_property_message(self, max_frequency))

    @property
    def global_temperature(self):
        """:obj:`float`: Simulation temperature in degrees Celsius."""
        return self.__stub.GetGlobalTemperature(self.msg).value

    @global_temperature.setter
    def global_temperature(self, global_temperature):
        self.__stub.SetGlobalTemperature(messages.double_property_message(self, global_temperature))

    @property
    def save_netlist(self):
        """:obj:`bool`: Flag indicating if the netlist output is saved."""
        return self.__stub.GetSaveNetlist(self.msg).value

    @save_netlist.setter
    def save_netlist(self, save_netlist):
        self.__stub.SetSaveNetlist(messages.bool_property_message(self, save_netlist))

    @property
    def netlist_export_spectre(self):
        """:obj:`bool`: Flag indicating if the netlist is exported in Spectre format."""
        return self.__stub.GetNetlistExportSpectre(self.msg).value

    @netlist_export_spectre.setter
    def netlist_export_spectre(self, netlist_export_spectre):
        self.__stub.SetNetlistExportSpectre(
            messages.bool_property_message(self, netlist_export_spectre)
        )

    @property
    def save_rfm(self):
        """:obj:`bool`: Flag indicating if an RFM file is exported."""
        return self.__stub.GetSaveRFM(self.msg).value

    @save_rfm.setter
    def save_rfm(self, save_rfm):
        self.__stub.SetSaveRFM(messages.bool_property_message(self, save_rfm))


class RaptorXAdvancedSettings(SimulationSettingsBase):
    """Represents advanced settings for RaptorX simulations."""

    __stub: RaptorXAdvancedSettingsServiceStub = StubAccessor(StubType.raptor_x_adv_sim_settings)

    @property
    def use_mesh_frequency(self):
        """:obj:`bool`: Flag indicating if the default meshing frequency is overridden."""
        return self.__stub.GetUseMeshFrequency(self.msg).value

    @use_mesh_frequency.setter
    def use_mesh_frequency(self, use_mesh_frequency):
        self.__stub.SetUseMeshFrequency(messages.bool_property_message(self, use_mesh_frequency))

    @property
    def mesh_frequency(self):
        """:obj:`str`: Mesh frequency override."""
        return self.__stub.GetMeshFrequency(self.msg).value

    @mesh_frequency.setter
    def mesh_frequency(self, mesh_frequency):
        self.__stub.SetMeshFrequency(messages.string_property_message(self, mesh_frequency))

    @property
    def use_edge_mesh(self):
        """:obj:`bool`: Flag indicating if the edge mesh is used."""
        return self.__stub.GetUseEdgeMesh(self.msg).value

    @use_edge_mesh.setter
    def use_edge_mesh(self, use_edge_mesh):
        self.__stub.SetUseEdgeMesh(messages.bool_property_message(self, use_edge_mesh))

    @property
    def edge_mesh(self):
        """:obj:`str`: Thickness and width of the exterior conductor filament."""
        return self.__stub.GetEdgeMesh(self.msg).value

    @edge_mesh.setter
    def edge_mesh(self, edge_mesh):
        self.__stub.SetEdgeMesh(messages.string_property_message(self, edge_mesh))

    @property
    def use_cells_per_wavelength(self):
        """:obj:`bool`: Flag indicating if cells per wavelength are used."""
        return self.__stub.GetUseCellsPerWavelength(self.msg).value

    @use_cells_per_wavelength.setter
    def use_cells_per_wavelength(self, use_cells_per_wavelength):
        self.__stub.SetUseCellsPerWavelength(
            messages.bool_property_message(self, use_cells_per_wavelength)
        )

    @property
    def cells_per_wavelength(self):
        """:obj:`int`: Number of cells that fit under each wavelength."""
        return self.__stub.GetCellsPerWavelength(self.msg).value

    @cells_per_wavelength.setter
    def cells_per_wavelength(self, cells_per_wavelength):
        self.__stub.SetCellsPerWavelength(
            messages.uint64_property_message(self, cells_per_wavelength)
        )

    @property
    def use_plane_projection_factor(self):
        """:obj:`bool`: Flag indicating if the plane projection factor is used."""
        return self.__stub.GetUsePlaneProjectionFactor(self.msg).value

    @use_plane_projection_factor.setter
    def use_plane_projection_factor(self, use_plane_projection_factor):
        self.__stub.SetUsePlaneProjectionFactor(
            messages.bool_property_message(self, use_plane_projection_factor)
        )

    @property
    def plane_projection_factor(self):
        """:obj:`float`: Plane projection factor for reducing the mesh complexity of large metal planes."""
        return self.__stub.GetPlaneProjectionFactor(self.msg).value

    @plane_projection_factor.setter
    def plane_projection_factor(self, plane_projection_factor):
        self.__stub.SetPlaneProjectionFactor(
            messages.double_property_message(self, plane_projection_factor)
        )

    @property
    def use_relaxed_z_axis(self):
        """:obj:`bool`: Flag indicating if simplified meshing is used along the z axis."""
        return self.__stub.GetUseRelaxedZAxis(self.msg).value

    @use_relaxed_z_axis.setter
    def use_relaxed_z_axis(self, use_relaxed_z_axis):
        self.__stub.SetUseRelaxedZAxis(messages.bool_property_message(self, use_relaxed_z_axis))

    @property
    def use_eliminate_slit_per_holes(self):
        """:obj:`bool`: Flag indicating if strain relief or thermal relief slits and holes are removed."""
        return self.__stub.GetUseEliminateSlitPerHoles(self.msg).value

    @use_eliminate_slit_per_holes.setter
    def use_eliminate_slit_per_holes(self, use_eliminate_slit_per_holes):
        self.__stub.SetUseEliminateSlitPerHoles(
            messages.bool_property_message(self, use_eliminate_slit_per_holes)
        )

    @property
    def eliminate_slit_per_holes(self):
        """:obj:`float`: Threshold for strain or thermal relief slits and hole polygon areas."""
        return self.__stub.GetEliminateSlitPerHoles(self.msg).value

    @eliminate_slit_per_holes.setter
    def eliminate_slit_per_holes(self, eliminate_slit_per_holes):
        self.__stub.SetEliminateSlitPerHoles(
            messages.double_property_message(self, eliminate_slit_per_holes)
        )

    @property
    def use_auto_removal_sliver_poly(self):
        """:obj:`bool`: Flag indicating if slight misaligned overlapping polygons are to be automatically aligned."""
        return self.__stub.GetUseAutoRemovalSliverPoly(self.msg).value

    @use_auto_removal_sliver_poly.setter
    def use_auto_removal_sliver_poly(self, use_auto_removal_sliver_poly):
        self.__stub.SetUseAutoRemovalSliverPoly(
            messages.bool_property_message(self, use_auto_removal_sliver_poly)
        )

    @property
    def auto_removal_sliver_poly(self):
        """:obj:`float`: Automatic sliver polygon removal tolerance."""
        return self.__stub.GetAutoRemovalSliverPoly(self.msg).value

    @auto_removal_sliver_poly.setter
    def auto_removal_sliver_poly(self, auto_removal_sliver_poly):
        self.__stub.SetAutoRemovalSliverPoly(
            messages.double_property_message(self, auto_removal_sliver_poly)
        )

    @property
    def use_accelerate_via_extraction(self):
        """:obj:`bool`: Flag indicating if neighboring vias are simplified/merged."""
        return self.__stub.GetUseAccelerateViaExtraction(self.msg).value

    @use_accelerate_via_extraction.setter
    def use_accelerate_via_extraction(self, use_accelerate_via_extraction):
        self.__stub.SetUseAccelerateViaExtraction(
            messages.bool_property_message(self, use_accelerate_via_extraction)
        )

    @property
    def use_enable_substrate_network_extraction(self):
        """:obj:`bool`: Flag indicating if modeling of substrate coupling effects \
        is enabled using equivalent distributed RC networks."""
        return self.__stub.GetUseEnableSubstrateNetworkExtraction(self.msg).value

    @use_enable_substrate_network_extraction.setter
    def use_enable_substrate_network_extraction(self, use_enable_substrate_network_extraction):
        self.__stub.SetUseEnableSubstrateNetworkExtraction(
            messages.bool_property_message(self, use_enable_substrate_network_extraction)
        )

    @property
    def use_lde(self):
        """:obj:`bool`: Flag indicating if variations in resistivity are taken into account."""
        return self.__stub.GetUseLDE(self.msg).value

    @use_lde.setter
    def use_lde(self, use_lde):
        self.__stub.SetUseLDE(messages.bool_property_message(self, use_lde))

    @property
    def use_extract_floating_metals_dummy(self):
        """:obj:`bool`: Flag indicating if floating metals are modeled as dummy fills."""
        return self.__stub.GetUseExtractFloatingMetalsDummy(self.msg).value

    @use_extract_floating_metals_dummy.setter
    def use_extract_floating_metals_dummy(self, use_extract_floating_metals_dummy):
        self.__stub.SetUseExtractFloatingMetalsDummy(
            messages.bool_property_message(self, use_extract_floating_metals_dummy)
        )

    @property
    def use_extract_floating_metals_floating(self):
        """:obj:`bool`: Flag indicating if floating metals are modeled as floating nets."""
        return self.__stub.GetUseExtractFloatingMetalsFloating(self.msg).value

    @use_extract_floating_metals_floating.setter
    def use_extract_floating_metals_floating(self, use_extract_floating_metals_floating):
        self.__stub.SetUseExtractFloatingMetalsFloating(
            messages.bool_property_message(self, use_extract_floating_metals_floating)
        )

    @property
    def use_enable_etch_transform(self):
        """:obj:`bool`: Flag indicating if layout is "pre-distorted" based on foundry rules."""
        return self.__stub.GetUseEnableEtchTransform(self.msg).value

    @use_enable_etch_transform.setter
    def use_enable_etch_transform(self, use_enable_etch_transform):
        self.__stub.SetUseEnableEtchTransform(
            messages.bool_property_message(self, use_enable_etch_transform)
        )

    @property
    def use_enable_hybrid_extraction(self):
        """:obj:`bool`: Flag indicating if the modeler is to split the layout into \
        two parts in an attempt to decrease the complexity."""
        return self.__stub.GetUseEnableHybridExtraction(self.msg).value

    @use_enable_hybrid_extraction.setter
    def use_enable_hybrid_extraction(self, use_enable_hybrid_extraction):
        self.__stub.SetUseEnableHybridExtraction(
            messages.bool_property_message(self, use_enable_hybrid_extraction)
        )

    @property
    def use_enable_advanced_cap_effects(self):
        """:obj:`bool`: Flag indicating if capacitance-related effects such as conformal dielectrics are applied."""
        return self.__stub.GetUseEnableAdvancedCapEffects(self.msg).value

    @use_enable_advanced_cap_effects.setter
    def use_enable_advanced_cap_effects(self, use_enable_advanced_cap_effects):
        self.__stub.SetUseEnableAdvancedCapEffects(
            messages.bool_property_message(self, use_enable_advanced_cap_effects)
        )

    @property
    def use_override_shrink_factor(self):
        """:obj:`bool`: Flag indicating if shrink factor is overridden."""
        return self.__stub.GetUseOverrideShrinkFac(self.msg).value

    @use_override_shrink_factor.setter
    def use_override_shrink_factor(self, use_override_shrink_factor):
        self.__stub.SetUseOverrideShrinkFac(
            messages.bool_property_message(self, use_override_shrink_factor)
        )

    @property
    def override_shrink_factor(self):
        """:obj:`float`: Shrink factor override value."""
        return self.__stub.GetOverrideShrinkFac(self.msg).value

    @override_shrink_factor.setter
    def override_shrink_factor(self, override_shrink_factor):
        self.__stub.SetOverrideShrinkFac(
            messages.double_property_message(self, override_shrink_factor)
        )

    @property
    def advanced_options(self):
        """:obj:`dict` { :obj:`str` : :obj:`list`[:obj:`str`] }: Advanced options."""
        return _to_options_dict(self.__stub.GetAdvancedOptions(self.msg))

    @advanced_options.setter
    def advanced_options(self, advanced_options):
        self.__stub.SetAdvancedOptions(
            _to_raptor_x_sim_settings_options_property_msg(self, advanced_options)
        )

    @property
    def net_settings_options(self):
        """:obj:`dict` { :obj:`str` : :obj:`list`[:obj:`str`] }: Options for net settings."""
        return _to_options_dict(self.__stub.GetNetSettingsOptions(self.msg))

    @net_settings_options.setter
    def net_settings_options(self, net_settings_options):
        self.__stub.SetNetSettingsOptions(
            _to_raptor_x_sim_settings_options_property_msg(self, net_settings_options)
        )
